# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import attempt_import

import itertools
import logging
from operator import attrgetter

from pyomo.common import DeveloperError
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.collections import ComponentMap
from pyomo.common.fileutils import Executable

from pyomo.contrib.cp import IntervalVar
from pyomo.contrib.cp.interval_var import (
    IntervalVarStartTime,
    IntervalVarEndTime,
    IntervalVarPresence,
    IntervalVarLength,
    ScalarIntervalVar,
    IntervalVarData,
    IndexedIntervalVar,
)
from pyomo.contrib.cp.sequence_var import (
    SequenceVar,
    ScalarSequenceVar,
    SequenceVarData,
)
from pyomo.contrib.cp.scheduling_expr.scheduling_logic import (
    AlternativeExpression,
    SpanExpression,
    SynchronizeExpression,
)
from pyomo.contrib.cp.scheduling_expr.precedence_expressions import (
    BeforeExpression,
    AtExpression,
)
from pyomo.contrib.cp.scheduling_expr.sequence_expressions import (
    NoOverlapExpression,
    FirstInSequenceExpression,
    LastInSequenceExpression,
    BeforeInSequenceExpression,
    PredecessorToExpression,
)
from pyomo.contrib.cp.scheduling_expr.step_function_expressions import (
    AlwaysIn,
    StepAt,
    StepAtStart,
    StepAtEnd,
    Pulse,
    CumulativeFunction,
    NegatedStepFunction,
)

from pyomo.core.base import (
    minimize,
    maximize,
    SortComponents,
    Block,
    Objective,
    Constraint,
    Var,
    Param,
    BooleanVar,
    LogicalConstraint,
    Suffix,
    value,
)
from pyomo.core.base.boolean_var import (
    ScalarBooleanVar,
    BooleanVarData,
    IndexedBooleanVar,
)
from pyomo.core.base.expression import ScalarExpression, ExpressionData
from pyomo.core.base.param import IndexedParam, ScalarParam, ParamData
from pyomo.core.base.var import ScalarVar, VarData, IndexedVar
import pyomo.core.expr as EXPR
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor, identify_variables
from pyomo.core.base import Set, RangeSet
from pyomo.core.base.set import SetProduct
from pyomo.repn.util import ExitNodeDispatcher
from pyomo.opt import WriterFactory, SolverFactory, TerminationCondition, SolverResults

### FIXME: Remove the following as soon as non-active components no
### longer report active==True
from pyomo.network import Port

###


def _finalize_docplex(module, available):
    if not available:
        return
    _deferred_element_getattr_dispatcher['start_time'] = module.start_of
    _deferred_element_getattr_dispatcher['end_time'] = module.end_of
    _deferred_element_getattr_dispatcher['length'] = module.length_of
    _deferred_element_getattr_dispatcher['is_present'] = module.presence_of

    # Scheduling dispatchers
    _before_dispatchers[_START_TIME, _START_TIME] = module.start_before_start
    _before_dispatchers[_START_TIME, _END_TIME] = module.start_before_end
    _before_dispatchers[_END_TIME, _START_TIME] = module.end_before_start
    _before_dispatchers[_END_TIME, _END_TIME] = module.end_before_end

    _at_dispatchers[_START_TIME, _START_TIME] = module.start_at_start
    _at_dispatchers[_START_TIME, _END_TIME] = module.start_at_end
    _at_dispatchers[_END_TIME, _START_TIME] = module.end_at_start
    _at_dispatchers[_END_TIME, _END_TIME] = module.end_at_end

    _time_point_dispatchers[_START_TIME] = module.start_of
    _time_point_dispatchers[_END_TIME] = module.end_of


cp, docplex_available = attempt_import('docplex.cp.model', callback=_finalize_docplex)
cp_solver, docplex_available = attempt_import('docplex.cp.solver')

logger = logging.getLogger('pyomo.contrib.cp')


# These are things that don't need special handling:
class _GENERAL:
    pass


# These are operations that need to be deferred sometimes, usually because of
# indirection:
class _START_TIME:
    pass


class _END_TIME:
    pass


class _DEFERRED_ELEMENT_CONSTRAINT:
    pass


class _ELEMENT_CONSTRAINT:
    pass


class _DEFERRED_BEFORE:
    pass


class _DEFERRED_AFTER:
    pass


class _DEFERRED_AT:
    pass


class _BEFORE:
    pass


class _AT:
    pass


class _IMPLIES:
    pass


class _LAND:
    pass


class _LOR:
    pass


class _XOR:
    pass


class _EQUIVALENT_TO:
    pass


def _check_var_domain(visitor, node, var):
    if not var.domain.isdiscrete():
        # Note: in the context of the current writer, this should be unreachable
        # because we can't handle non-discrete variables at all, so there will
        # already be errors handling the children of this expression.
        raise ValueError(
            "Variable indirection '%s' contains argument '%s', "
            "which is not a discrete variable" % (node, var)
        )
    bnds = var.bounds
    if None in bnds:
        raise ValueError(
            "Variable indirection '%s' contains argument '%s', "
            "which is not restricted to a finite discrete domain" % (node, var)
        )
    return var.domain & RangeSet(*bnds)


def _handle_getitem(visitor, node, *data):
    # First we need to determine the range for each of the the
    # arguments.  They can be:
    #
    #  - simple values
    #  - docplex integer variables
    #  - docplex integer expressions
    arg_domain = []
    arg_scale = []
    expr = 0
    mult = 1
    # Note: skipping the first argument: that should be the IndexedComponent
    for i, arg in enumerate(data[1:]):
        if arg[1].__class__ in EXPR.native_types:
            arg_set = Set(initialize=[arg[1]])
            arg_set.construct()
            arg_domain.append(arg_set)
            arg_scale.append(None)
        elif node.arg(i + 1).is_expression_type():
            # This argument is an expression.  It could be any
            # combination of any number of integer variables, as long as
            # the resulting expression is still an IntExpression.  We
            # can't really rely on FBBT here, because we need to know
            # that the expression returns values in a regular domain
            # (i.e., the set of possible values has to have a start,
            # end, and finite, regular step).
            #
            # We will brute force it: go through every combination of
            # every variable and record the resulting expression value.
            arg_expr = node.arg(i + 1)
            var_list = list(identify_variables(arg_expr, include_fixed=False))
            var_domain = [list(_check_var_domain(visitor, node, v)) for v in var_list]
            arg_vals = set()
            for var_vals in itertools.product(*var_domain):
                for v, val in zip(var_list, var_vals):
                    v.set_value(val)
                arg_vals.add(arg_expr())
            # Now that we have all the values that define the domain of
            # the result of the expression, stick them into a set and
            # rely on the Set infrastructure to calculate (and verify)
            # the interval.
            arg_set = Set(initialize=sorted(arg_vals))
            arg_set.construct()
            interval = arg_set.get_interval()
            if not interval[2]:
                raise ValueError(
                    "Variable indirection '%s' contains argument expression "
                    "'%s' that does not evaluate to a simple discrete set"
                    % (node, arg_expr)
                )
            arg_domain.append(arg_set)
            arg_scale.append(interval)
        else:
            # This had better be a simple variable over a regular
            # discrete domain.  When we add support for categorical
            # variables, we will need to ensure that the categoricals
            # have already been converted to simple integer domains by
            # this point.
            var = node.arg(i + 1)
            arg_domain.append(_check_var_domain(visitor, node, var))
            arg_scale.append(arg_domain[-1].get_interval())
        # Build the expression that maps arguments to GetItem() to a
        # position in the elements list
        if arg_scale[-1] is not None:
            _min, _max, _step = arg_scale[-1]
            # ESJ: Have to use integer division here because otherwise, later,
            # when we construct the element constraint, docplex won't believe
            # the index is an integer expression.
            if _step is None:
                raise ValueError(
                    "Variable indirection '%s' is over a discrete domain "
                    "without a constant step size. This is not supported." % node
                )
            expr += mult * (arg[1] - _min) // _step
            # This could be (_max - _min) // _step + 1, but that assumes
            # that the set correctly collapsed the bounds and that the
            # lower and upper bounds were part of the step.  That
            # *should* be the case for Set, but I am suffering from a
            # crisis of confidence at the moment.
            mult *= len(arg_domain[-1])
    # Get the list of all elements selectable by the argument
    # expression(s); fill in new variables for any indices allowable by
    # the argument expression(s) but not present in the IndexedComponent
    # indexing set.
    elements = []
    for idx in SetProduct(*arg_domain):
        try:
            idx = idx if len(idx) > 1 else idx[0]
            elements.append(data[0][1][idx])
        except KeyError:
            raise ValueError(
                "Variable indirection '%s' permits an index '%s' "
                "that is not a valid key. In CP Optimizer, this is a "
                "structural infeasibility." % (node, idx)
            )
            # NOTE: If we thought it was the right thing to do in the future, we
            # could fill in with a bogus variable and add a constraint
            # disallowing it from being selected
    try:
        return (_ELEMENT_CONSTRAINT, cp.element(elements, expr))
    except AssertionError:
        return (_DEFERRED_ELEMENT_CONSTRAINT, (elements, expr))


_element_constraint_attr_dispatcher = {
    'before': _DEFERRED_BEFORE,
    'after': _DEFERRED_AFTER,
    'at': _DEFERRED_AT,
    'implies': _IMPLIES,
    'land': _LAND,
    'lor': _LOR,
    'xor': _XOR,
    'equivalent_to': _EQUIVALENT_TO,
}
# This will get populated when cp is finally imported
_deferred_element_getattr_dispatcher = {}


def _handle_getattr(visitor, node, obj, attr):
    # We either end up here because we do not yet know the list of variables to
    # make an element constraint (the first case) or because we are asking for
    # an attribute on something with indirection, so at this point we *have* a
    # constructed element constraint (the second case).
    if obj[0] is _DEFERRED_ELEMENT_CONSTRAINT:
        # then obj[1] is a list of cp thingies that we need to get the attr on,
        # and then at the end we need to make the element constraint we couldn't
        # make before.
        try:
            ans = list(map(_deferred_element_getattr_dispatcher[attr[1]], obj[1][0]))
        except KeyError:
            logger.error("Unrecognized attribute in GetAttrExpression: %s." % attr[1])
            raise
        return (_ELEMENT_CONSTRAINT, cp.element(array=ans, index=obj[1][1]))
    elif obj[0] is _ELEMENT_CONSTRAINT:
        try:
            return (_element_constraint_attr_dispatcher[attr[1]], obj)
        except KeyError:
            logger.error(
                "Unrecognized attribute in GetAttrExpression:"
                "%s. Found for object: %s" % (attr[1], obj[1])
            )
            raise
    else:
        raise DeveloperError(
            "Unrecognized argument type '%s' to getattr dispatcher." % obj[0]
        )


def _before_boolean_var(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        if child.fixed:
            return False, (_GENERAL, child.value)
        nm = child.name if visitor.symbolic_solver_labels else None
        # Sorry, universe, but docplex doesn't know the difference between
        # Boolean and Binary...
        cpx_var = cp.binary_var(name=nm)
        # Because I want to pretend the world is sane from here on out, we will
        # return a Boolean expression (in docplex land) so this can be used as
        # an argument to logical expressions later
        visitor.var_map[_id] = cpx_var == 1
        visitor.pyomo_to_docplex[child] = cpx_var
    return False, (_GENERAL, visitor.var_map[_id])


def _before_indexed_boolean_var(visitor, child):
    cpx_vars = {}
    for i, v in child.items():
        if v.fixed:
            cpx_vars[i] = v.value
            continue
        cpx_var = cp.binary_var(name=v.name if visitor.symbolic_solver_labels else None)
        visitor.cpx.add(cpx_var)
        visitor.var_map[id(v)] = cpx_var == 1
        visitor.pyomo_to_docplex[v] = cpx_var
        cpx_vars[i] = cpx_var == 1
    return False, (_GENERAL, cpx_vars)


def _before_param(visitor, child):
    return False, (_GENERAL, value(child))


def _before_indexed_param(visitor, child):
    return False, (_GENERAL, {idx: value(p) for idx, p in child.items()})


def _create_docplex_var(pyomo_var, name=None):
    if pyomo_var.is_binary():
        return cp.binary_var(name=name)
    elif pyomo_var.is_integer():
        return cp.integer_var(
            min=pyomo_var.bounds[0], max=pyomo_var.bounds[1], name=name
        )
    elif pyomo_var.domain.isdiscrete():
        if pyomo_var.domain.isfinite():
            return cp.integer_var(domain=[d for d in pyomo_var.domain], name=name)
        else:
            # If we ever want to handle this case, I think we might be able to
            # make a normal integer var and then constrain it into the
            # domain. But no reason to go to the effort for now because I don't
            # know if the solver can even work with such a var.
            raise ValueError(
                "The LogicalToDoCplex writer does not support "
                "infinite discrete domains. Cannot write "
                "Var '%s' with domain '%s'" % (pyomo_var.name, pyomo_var.domain)
            )
    else:
        raise ValueError(
            "The LogicalToDoCplex writer can only support "
            "integer- or Boolean-valued variables. Cannot "
            "write Var '%s' with domain '%s'" % (pyomo_var.name, pyomo_var.domain)
        )


def _before_var(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        if child.fixed:
            return False, (_GENERAL, child.value)
        cpx_var = _create_docplex_var(
            child, name=child.name if visitor.symbolic_solver_labels else None
        )
        visitor.cpx.add(cpx_var)
        visitor.var_map[_id] = cpx_var
        visitor.pyomo_to_docplex[child] = cpx_var
    return False, (_GENERAL, visitor.var_map[_id])


def _before_indexed_var(visitor, child):
    cpx_vars = {}
    for i, v in child.items():
        cpx_var = _create_docplex_var(
            v, name=v.name if visitor.symbolic_solver_labels else None
        )
        visitor.cpx.add(cpx_var)
        visitor.var_map[id(v)] = cpx_var
        visitor.pyomo_to_docplex[v] = cpx_var
        cpx_vars[i] = cpx_var
    return False, (_GENERAL, cpx_vars)


def _handle_named_expression_node(visitor, node, expr):
    visitor._named_expressions[id(node)] = expr[1]
    return expr


def _before_named_expression(visitor, child):
    _id = id(child)
    if _id not in visitor._named_expressions:
        return True, None
    return False, (_GENERAL, visitor._named_expressions[_id])


def _create_docplex_interval_var(visitor, interval_var):
    # Create a new docplex interval var and then figure out all the info that
    # gets stored on it
    nm = interval_var.name if visitor.symbolic_solver_labels else None
    cpx_interval_var = cp.interval_var(name=nm)
    visitor.var_map[id(interval_var)] = cpx_interval_var
    visitor.pyomo_to_docplex[interval_var] = cpx_interval_var

    # Figure out if it exists
    if interval_var.is_present.fixed and not interval_var.is_present.value:
        # Someone has fixed that this will not get scheduled.
        cpx_interval_var.set_absent()
    elif interval_var.optional:
        cpx_interval_var.set_optional()
    else:
        cpx_interval_var.set_present()

    # Figure out constraints on its length
    length = interval_var.length
    if length.fixed:
        cpx_interval_var.set_length(length.value)
    if length.lb is not None:
        cpx_interval_var.set_length_min(length.lb)
    if length.ub is not None:
        cpx_interval_var.set_length_max(length.ub)

    # Figure out constraints on start time
    start_time = interval_var.start_time
    if start_time.fixed:
        cpx_interval_var.set_start(start_time.value)
    else:
        if start_time.lb is not None:
            cpx_interval_var.set_start_min(start_time.lb)
        if start_time.ub is not None:
            cpx_interval_var.set_start_max(start_time.ub)

    # Figure out constraints on end time
    end_time = interval_var.end_time
    if end_time.fixed:
        cpx_interval_var.set_end(end_time.value)
    else:
        if end_time.lb is not None:
            cpx_interval_var.set_end_min(end_time.lb)
        if end_time.ub is not None:
            cpx_interval_var.set_end_max(end_time.ub)

    return cpx_interval_var


def _create_docplex_sequence_var(visitor, sequence_var):
    nm = sequence_var.name if visitor.symbolic_solver_labels else None

    cpx_seq_var = cp.sequence_var(
        name=nm,
        vars=[
            _get_docplex_interval_var(visitor, v) for v in sequence_var.interval_vars
        ],
    )
    visitor.var_map[id(sequence_var)] = cpx_seq_var
    return cpx_seq_var


def _get_docplex_interval_var(visitor, interval_var):
    # We might already have the interval_var and just need to retrieve it
    if id(interval_var) in visitor.var_map:
        cpx_interval_var = visitor.var_map[id(interval_var)]
    else:
        cpx_interval_var = _create_docplex_interval_var(visitor, interval_var)
        visitor.cpx.add(cpx_interval_var)
    return cpx_interval_var


def _get_docplex_sequence_var(visitor, sequence_var):
    if id(sequence_var) in visitor.var_map:
        cpx_seq_var = visitor.var_map[id(sequence_var)]
    else:
        cpx_seq_var = _create_docplex_sequence_var(visitor, sequence_var)
        visitor.cpx.add(cpx_seq_var)
    return cpx_seq_var


def _before_sequence_var(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        cpx_seq_var = _get_docplex_sequence_var(visitor, child)
        visitor.var_map[_id] = cpx_seq_var
        visitor.pyomo_to_docplex[child] = cpx_seq_var

    return False, (_GENERAL, visitor.var_map[_id])


def _before_interval_var(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        cpx_interval_var = _get_docplex_interval_var(visitor, child)
        visitor.var_map[_id] = cpx_interval_var
        visitor.pyomo_to_docplex[child] = cpx_interval_var

    return False, (_GENERAL, visitor.var_map[_id])


def _before_indexed_interval_var(visitor, child):
    cpx_vars = {}
    for i, v in child.items():
        cpx_interval_var = _get_docplex_interval_var(visitor, v)
        visitor.var_map[id(v)] = cpx_interval_var
        visitor.pyomo_to_docplex[v] = cpx_interval_var
        cpx_vars[i] = cpx_interval_var
    return False, (_GENERAL, cpx_vars)


def _before_interval_var_start_time(visitor, child):
    _id = id(child)
    interval_var = child.get_associated_interval_var()
    if _id not in visitor.var_map:
        cpx_interval_var = _get_docplex_interval_var(visitor, interval_var)

    return False, (_START_TIME, visitor.var_map[id(interval_var)])


def _before_interval_var_end_time(visitor, child):
    _id = id(child)
    interval_var = child.get_associated_interval_var()
    if _id not in visitor.var_map:
        cpx_interval_var = _get_docplex_interval_var(visitor, interval_var)

    return False, (_END_TIME, visitor.var_map[id(interval_var)])


def _before_interval_var_length(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        interval_var = child.get_associated_interval_var()
        cpx_interval_var = _get_docplex_interval_var(visitor, interval_var)

        visitor.var_map[_id] = cp.length_of(cpx_interval_var)
    # There aren't any special types of constraints involving the length, so we
    # just treat this expression as if it's a normal variable.
    return False, (_GENERAL, visitor.var_map[_id])


def _before_interval_var_presence(visitor, child):
    _id = id(child)
    if _id not in visitor.var_map:
        interval_var = child.get_associated_interval_var()
        cpx_interval_var = _get_docplex_interval_var(visitor, interval_var)

        visitor.var_map[_id] = cp.presence_of(cpx_interval_var)
    # There aren't any special types of constraints involving the presence, so
    # we just treat this expression as if it's a normal variable.
    return False, (_GENERAL, visitor.var_map[_id])


def _handle_step_at_node(visitor, node):
    return False, (_GENERAL, cp.step_at(node._time, node._height))


def _handle_step_at_start_node(visitor, node):
    cpx_var = _get_docplex_interval_var(visitor, node._time)
    return False, (_GENERAL, cp.step_at_start(cpx_var, node._height))


def _handle_step_at_end_node(visitor, node):
    cpx_var = _get_docplex_interval_var(visitor, node._time)
    return False, (_GENERAL, cp.step_at_end(cpx_var, node._height))


def _handle_pulse_node(visitor, node):
    cpx_var = _get_docplex_interval_var(visitor, node._interval_var)
    return False, (_GENERAL, cp.pulse(cpx_var, node._height))


def _handle_negated_step_function_node(visitor, node):
    return _step_function_handles[node.args[0].__class__](visitor, node.args[0])


def _handle_cumulative_function(visitor, node):
    expr = 0
    for arg in node.args:
        if arg.__class__ is NegatedStepFunction:
            expr -= _handle_negated_step_function_node(visitor, arg)[1][1]
        else:
            expr += _step_function_handles[arg.__class__](visitor, arg)[1][1]

    return False, (_GENERAL, expr)


_step_function_handles = {
    StepAt: _handle_step_at_node,
    StepAtStart: _handle_step_at_start_node,
    StepAtEnd: _handle_step_at_end_node,
    Pulse: _handle_pulse_node,
    CumulativeFunction: _handle_cumulative_function,
    NegatedStepFunction: _handle_negated_step_function_node,
}
step_func_expression_types = _step_function_handles.keys()

##
# Algebraic expressions
##


def _get_int_valued_expr(arg):
    if arg[0] in {_GENERAL, _ELEMENT_CONSTRAINT}:
        return arg[1]
    elif arg[0] is _START_TIME:
        return cp.start_of(arg[1])
    elif arg[0] is _END_TIME:
        return cp.end_of(arg[1])
    else:
        raise DeveloperError(
            "Attempting to get a docplex integer-valued "
            "expression from object in class %s" % str(arg[0])
        )


def _get_bool_valued_expr(arg):
    if arg[0] is _GENERAL:
        return arg[1]
    elif arg[0] is _ELEMENT_CONSTRAINT:
        # docplex doesn't bother to check if 'element' expressions are integer-
        # or boolean-valued: they just complain if you use them in a boolean
        # context. So if we are about to use one that way, we set it equivalent
        # to True so that it will be boolean-valued according to docplex's
        # idiosyncrasies.
        return arg[1] == True
    elif arg[0] is _BEFORE:
        # We're using a start-before-start or its ilk in a boolean-valued
        # context. docplex doesn't believe these things are boolean-valued, so
        # we have to convert to the inequality version:
        lhs, rhs = arg[2]
        return _handle_inequality_node(None, None, lhs, rhs)[1]
    elif arg[0] is _AT:
        # Same as above, but now we need an equality node
        lhs, rhs = arg[2]
        return _handle_equality_node(None, None, lhs, rhs)[1]
    else:
        raise DeveloperError(
            "Attempting to get a docplex Boolean-valued "
            "expression from object in class %s" % str(arg[0])
        )


def _handle_monomial_expr(visitor, node, arg1, arg2):
    # Monomial terms show up a lot.  This handles some common
    # simplifications (necessary in part for the unit tests)
    if arg2[1].__class__ in EXPR.native_types:
        return _GENERAL, arg1[1] * arg2[1]
    elif arg1[1].__class__ in EXPR.native_types and arg1[1] == 1:
        return arg2
    return (_GENERAL, cp.times(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)))


def _handle_sum_node(visitor, node, *args):
    return (
        _GENERAL,
        sum(
            (_get_int_valued_expr(arg) for arg in args[1:]),
            _get_int_valued_expr(args[0]),
        ),
    )


def _handle_negation_node(visitor, node, arg1):
    return (_GENERAL, cp.times(-1, _get_int_valued_expr(arg1)))


def _handle_product_node(visitor, node, arg1, arg2):
    return (_GENERAL, cp.times(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)))


def _handle_division_node(visitor, node, arg1, arg2):
    return (
        _GENERAL,
        cp.float_div(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)),
    )


def _handle_pow_node(visitor, node, arg1, arg2):
    return (_GENERAL, cp.power(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)))


def _handle_abs_node(visitor, node, arg1):
    return (_GENERAL, cp.abs(_get_int_valued_expr(arg1)))


def _handle_min_node(visitor, node, *args):
    return (_GENERAL, cp.min((_get_int_valued_expr(arg) for arg in args)))


def _handle_max_node(visitor, node, *args):
    return (_GENERAL, cp.max((_get_int_valued_expr(arg) for arg in args)))


##
# Relational expressions
##


def _handle_equality_node(visitor, node, arg1, arg2):
    return (_GENERAL, cp.equal(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)))


def _handle_inequality_node(visitor, node, arg1, arg2):
    return (
        _GENERAL,
        cp.less_or_equal(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)),
    )


def _handle_ranged_inequality_node(visitor, node, arg1, arg2, arg3):
    return (
        _GENERAL,
        cp.range(
            _get_int_valued_expr(arg2),
            lb=_get_int_valued_expr(arg1),
            ub=_get_int_valued_expr(arg3),
        ),
    )


def _handle_not_equal_node(visitor, node, arg1, arg2):
    return (_GENERAL, cp.diff(_get_int_valued_expr(arg1), _get_int_valued_expr(arg2)))


##
# Logical expressions
##


def _handle_and_node(visitor, node, *args):
    return (_GENERAL, cp.logical_and((_get_bool_valued_expr(arg) for arg in args)))


def _handle_or_node(visitor, node, *args):
    return (_GENERAL, cp.logical_or((_get_bool_valued_expr(arg) for arg in args)))


def _handle_xor_node(visitor, node, arg1, arg2):
    return (
        _GENERAL,
        cp.equal(
            cp.count([_get_bool_valued_expr(arg1), _get_bool_valued_expr(arg2)], 1), 1
        ),
    )


def _handle_not_node(visitor, node, arg):
    return (_GENERAL, cp.logical_not(_get_bool_valued_expr(arg)))


def _handle_equivalence_node(visitor, node, arg1, arg2):
    return (
        _GENERAL,
        cp.equal(_get_bool_valued_expr(arg1), _get_bool_valued_expr(arg2)),
    )


def _handle_implication_node(visitor, node, arg1, arg2):
    return (
        _GENERAL,
        cp.if_then(_get_bool_valued_expr(arg1), _get_bool_valued_expr(arg2)),
    )


def _handle_exactly_node(visitor, node, *args):
    return (
        _GENERAL,
        cp.equal(
            cp.count((_get_bool_valued_expr(arg) for arg in args[1:]), 1),
            _get_int_valued_expr(args[0]),
        ),
    )


def _handle_at_most_node(visitor, node, *args):
    return (
        _GENERAL,
        cp.less_or_equal(
            cp.count((_get_bool_valued_expr(arg) for arg in args[1:]), 1),
            _get_int_valued_expr(args[0]),
        ),
    )


def _handle_at_least_node(visitor, node, *args):
    return (
        _GENERAL,
        cp.greater_or_equal(
            cp.count((_get_bool_valued_expr(arg) for arg in args[1:]), 1),
            _get_int_valued_expr(args[0]),
        ),
    )


def _handle_all_diff_node(visitor, node, *args):
    return (_GENERAL, cp.all_diff(_get_int_valued_expr(arg) for arg in args))


def _handle_count_if_node(visitor, node, *args):
    return (_GENERAL, cp.count((_get_bool_valued_expr(arg) for arg in args), 1))


## CallExpression handllers


def _before_call_dispatcher(visitor, node, *args):
    if len(args) == 2:
        return _handle_inequality_node(visitor, node, args[0], args[1])
    else:  # a delay is also specified
        lhs = _handle_sum_node(visitor, node, args[0], args[2])
        return _handle_inequality_node(visitor, node, lhs, args[1])


def _after_call_dispatcher(visitor, node, *args):
    if len(args) == 2:
        return _handle_inequality_node(visitor, node, args[1], args[0])
    else:  # delay is also specified
        lhs = _handle_sum_node(visitor, node, args[1], args[2])
        return _handle_inequality_node(visitor, node, lhs, args[0])


def _at_call_dispatcher(visitor, node, *args):
    if len(args) == 2:
        return _handle_equality_node(visitor, node, args[0], args[1])
    else:  # a delay is also specified
        rhs = _handle_sum_node(visitor, node, args[1], args[2])
        return _handle_equality_node(visitor, node, args[0], rhs)


_call_dispatchers = {
    _DEFERRED_BEFORE: _before_call_dispatcher,
    _DEFERRED_AFTER: _after_call_dispatcher,
    _DEFERRED_AT: _at_call_dispatcher,
    _IMPLIES: _handle_implication_node,
    _LAND: _handle_and_node,
    _LOR: _handle_or_node,
    _XOR: _handle_xor_node,
    _EQUIVALENT_TO: _handle_equivalence_node,
}


def _handle_call(visitor, node, *args):
    return _call_dispatchers[args[0][0]](visitor, node, args[0][1], *args[1:])


##
# Scheduling
##

# This will get populated when cp is finally imported
_before_dispatchers = {}
_at_dispatchers = {}
_time_point_dispatchers = {_GENERAL: lambda x: x, _ELEMENT_CONSTRAINT: lambda x: x}

_non_precedence_types = {_GENERAL, _ELEMENT_CONSTRAINT}


def _handle_before_expression_node(visitor, node, time1, time2, delay):
    t1 = (_GENERAL, _time_point_dispatchers[time1[0]](time1[1]))
    t2 = (_GENERAL, _time_point_dispatchers[time2[0]](time2[1]))
    lhs = _handle_sum_node(visitor, None, t1, delay)
    if time1[0] in _non_precedence_types or time2[0] in _non_precedence_types:
        # we already know we can't use a start_before_start function or its ilk:
        # Just build the correct inequality.
        return _handle_inequality_node(visitor, None, lhs, t2)

    # If this turns out to be the root, we can use the second return, but we
    # also pass the args for the inequality expression in case we use this in a
    # boolean-valued context.
    return (
        _BEFORE,
        _before_dispatchers[time1[0], time2[0]](time1[1], time2[1], delay[1]),
        (lhs, t2),
    )


def _handle_at_expression_node(visitor, node, time1, time2, delay):
    t1 = (_GENERAL, _time_point_dispatchers[time1[0]](time1[1]))
    t2 = (_GENERAL, _time_point_dispatchers[time2[0]](time2[1]))
    lhs = _handle_sum_node(visitor, None, t1, delay)
    if time1[0] in _non_precedence_types or time2[0] in _non_precedence_types:
        # we can't use a start_before_start function or its ilk: Just build the
        # correct inequality.
        return _handle_equality_node(visitor, None, lhs, t2)

    return (
        _AT,
        _at_dispatchers[time1[0], time2[0]](time1[1], time2[1], delay[1]),
        (lhs, t2),
    )


def _handle_always_in_node(visitor, node, cumul_func, lb, ub, start, end):
    return (
        _GENERAL,
        cp.always_in(cumul_func[1], interval=(start[1], end[1]), min=lb[1], max=ub[1]),
    )


def _handle_no_overlap_expression_node(visitor, node, seq_var):
    return _GENERAL, cp.no_overlap(seq_var[1])


def _handle_first_in_sequence_expression_node(visitor, node, interval_var, seq_var):
    return _GENERAL, cp.first(seq_var[1], interval_var[1])


def _handle_last_in_sequence_expression_node(visitor, node, interval_var, seq_var):
    return _GENERAL, cp.last(seq_var[1], interval_var[1])


def _handle_before_in_sequence_expression_node(
    visitor, node, before_var, after_var, seq_var
):
    return _GENERAL, cp.before(seq_var[1], before_var[1], after_var[1])


def _handle_predecessor_to_expression_node(
    visitor, node, before_var, after_var, seq_var
):
    return _GENERAL, cp.previous(seq_var[1], before_var[1], after_var[1])


def _handle_span_expression_node(visitor, node, *args):
    return _GENERAL, cp.span(args[0][1], [arg[1] for arg in args[1:]])


def _handle_alternative_expression_node(visitor, node, *args):
    return _GENERAL, cp.alternative(args[0][1], [arg[1] for arg in args[1:]])


def _handle_synchronize_expression_node(visitor, node, *args):
    return _GENERAL, cp.synchronize(args[0][1], [arg[1] for arg in args[1:]])


_operator_handles = {
    EXPR.GetItemExpression: _handle_getitem,
    EXPR.GetAttrExpression: _handle_getattr,
    EXPR.CallExpression: _handle_call,
    EXPR.NegationExpression: _handle_negation_node,
    EXPR.ProductExpression: _handle_product_node,
    EXPR.DivisionExpression: _handle_division_node,
    EXPR.PowExpression: _handle_pow_node,
    EXPR.AbsExpression: _handle_abs_node,
    EXPR.MonomialTermExpression: _handle_monomial_expr,
    EXPR.SumExpression: _handle_sum_node,
    EXPR.MinExpression: _handle_min_node,
    EXPR.MaxExpression: _handle_max_node,
    EXPR.NotExpression: _handle_not_node,
    EXPR.EquivalenceExpression: _handle_equivalence_node,
    EXPR.ImplicationExpression: _handle_implication_node,
    EXPR.AndExpression: _handle_and_node,
    EXPR.OrExpression: _handle_or_node,
    EXPR.XorExpression: _handle_xor_node,
    EXPR.ExactlyExpression: _handle_exactly_node,
    EXPR.AtMostExpression: _handle_at_most_node,
    EXPR.AtLeastExpression: _handle_at_least_node,
    EXPR.AllDifferentExpression: _handle_all_diff_node,
    EXPR.CountIfExpression: _handle_count_if_node,
    EXPR.EqualityExpression: _handle_equality_node,
    EXPR.NotEqualExpression: _handle_not_equal_node,
    EXPR.InequalityExpression: _handle_inequality_node,
    EXPR.RangedExpression: _handle_ranged_inequality_node,
    BeforeExpression: _handle_before_expression_node,
    AtExpression: _handle_at_expression_node,
    AlwaysIn: _handle_always_in_node,
    ExpressionData: _handle_named_expression_node,
    ScalarExpression: _handle_named_expression_node,
    NoOverlapExpression: _handle_no_overlap_expression_node,
    FirstInSequenceExpression: _handle_first_in_sequence_expression_node,
    LastInSequenceExpression: _handle_last_in_sequence_expression_node,
    BeforeInSequenceExpression: _handle_before_in_sequence_expression_node,
    PredecessorToExpression: _handle_predecessor_to_expression_node,
    SpanExpression: _handle_span_expression_node,
    AlternativeExpression: _handle_alternative_expression_node,
    SynchronizeExpression: _handle_synchronize_expression_node,
}


class LogicalToDoCplex(StreamBasedExpressionVisitor):
    exit_node_dispatcher = ExitNodeDispatcher(_operator_handles)
    # NOTE: Because of indirection, we can encounter indexed Params and Vars in
    # expressions

    _var_handles = {
        IntervalVarStartTime: _before_interval_var_start_time,
        IntervalVarEndTime: _before_interval_var_end_time,
        IntervalVarLength: _before_interval_var_length,
        IntervalVarPresence: _before_interval_var_presence,
        ScalarIntervalVar: _before_interval_var,
        IntervalVarData: _before_interval_var,
        IndexedIntervalVar: _before_indexed_interval_var,
        ScalarSequenceVar: _before_sequence_var,
        SequenceVarData: _before_sequence_var,
        ScalarVar: _before_var,
        VarData: _before_var,
        IndexedVar: _before_indexed_var,
        ScalarBooleanVar: _before_boolean_var,
        BooleanVarData: _before_boolean_var,
        IndexedBooleanVar: _before_indexed_boolean_var,
        ExpressionData: _before_named_expression,
        ScalarExpression: _before_named_expression,
        IndexedParam: _before_indexed_param,
        ScalarParam: _before_param,
        ParamData: _before_param,
    }

    def __init__(self, cpx_model, symbolic_solver_labels=False):
        self.cpx = cpx_model
        self.symbolic_solver_labels = symbolic_solver_labels
        self._process_node = self._process_node_bx

        self.var_map = {}
        self._named_expressions = {}
        self.pyomo_to_docplex = ComponentMap()

    def initializeWalker(self, expr):
        expr, src, src_idx = expr
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, result
        return True, expr

    def beforeChild(self, node, child, child_idx):
        # Return native types
        if child.__class__ in EXPR.native_types:
            return False, (_GENERAL, child)

        if child.__class__ in step_func_expression_types:
            return _step_function_handles[child.__class__](self, child)

        # Convert Vars Logical vars to docplex equivalents
        if not child.is_expression_type() or child.is_named_expression_type():
            return self._var_handles[child.__class__](self, child)

        return True, None

    def exitNode(self, node, data):
        return self.exit_node_dispatcher[node.__class__](self, node, *data)

    finalizeResult = None


# [ESJ 11/7/22]: TODO: We should revisit this method in the future, as it is not
# very efficient.
def collect_valid_components(model, active=True, sort=None, valid=set(), targets=set()):
    assert active in (True, None)
    unrecognized = {}
    components = {k: [] for k in targets}
    for obj in model.component_data_objects(active=True, descend_into=True, sort=sort):
        # HACK around #3045
        if not hasattr(obj, 'ctype'):
            continue
        ctype = obj.ctype
        if ctype in components:
            components[ctype].append(obj)
        elif ctype not in valid:
            if ctype not in unrecognized:
                unrecognized[ctype] = [obj]
            else:
                unrecognized[ctype].append(obj)

    return components, unrecognized


@WriterFactory.register(
    'docplex_model', 'Generate the corresponding docplex model object'
)
class DocplexWriter:
    CONFIG = ConfigDict('docplex_model_writer')
    CONFIG.declare(
        'symbolic_solver_labels',
        ConfigValue(
            default=False,
            domain=bool,
            description='Write Pyomo Var and Constraint names to docplex model',
        ),
    )

    def __init__(self):
        self.config = self.CONFIG()

    def write(self, model, **options):
        config = options.pop('config', self.config)(options)

        components, unknown = collect_valid_components(
            model,
            active=True,
            sort=SortComponents.deterministic,
            valid={
                Block,
                Objective,
                Constraint,
                Var,
                Param,
                BooleanVar,
                LogicalConstraint,
                Suffix,
                # FIXME: Non-active components should not report as Active
                Set,
                RangeSet,
                Port,
            },
            targets={
                Objective,
                Constraint,
                LogicalConstraint,
                IntervalVar,
                SequenceVar,
            },
        )
        if unknown:
            raise ValueError(
                "The model ('%s') contains the following active components "
                "that the docplex writer does not know how to process:\n\t%s"
                % (
                    model.name,
                    "\n\t".join(
                        "%s:\n\t\t%s" % (k, "\n\t\t".join(map(attrgetter('name'), v)))
                        for k, v in unknown.items()
                    ),
                )
            )

        cpx_model = cp.CpoModel()
        visitor = LogicalToDoCplex(
            cpx_model, symbolic_solver_labels=config.symbolic_solver_labels
        )

        active_objs = components[Objective]
        # [ESJ 09/29/22]: TODO: I think that CP Optimizer can support
        # multiple objectives. We should generalize this later, but for
        # now I don't much care.
        if len(active_objs) > 1:
            raise ValueError(
                "More than one active objective defined for "
                "input model '%s': Cannot write to docplex." % model.name
            )
        elif len(active_objs) == 1:
            obj = active_objs[0]
            obj_expr = visitor.walk_expression((obj.expr, obj, 0))
            if obj.sense is minimize:
                cpx_model.add(cp.minimize(obj_expr[1]))
            else:
                cpx_model.add(cp.maximize(obj_expr[1]))

        # No objective is fine too, this is CP afterall...

        # Write algebraic constraints
        for cons in components[Constraint]:
            expr = visitor.walk_expression((cons.body, cons, 0))
            if cons.lower is not None and cons.upper is not None:
                cpx_model.add(cp.range(expr[1], lb=cons.lb, ub=cons.ub))
            elif cons.lower is not None:
                cpx_model.add(cons.lb <= expr[1])
            elif cons.upper is not None:
                cpx_model.add(cons.ub >= expr[1])

        # Write interval vars (these are secretly constraints if they have to be
        # scheduled)
        for var in components[IntervalVar]:
            # we just walk it so it gets added to the model. Note that
            # adding it again here would add it for a second time, so that's
            # why we don't.
            visitor.walk_expression((var, var, 0))

        # Write logical constraints
        for cons in components[LogicalConstraint]:
            expr = visitor.walk_expression((cons.expr, cons, 0))
            if expr[0] is _ELEMENT_CONSTRAINT:
                # Make the expression into a docplex-approved boolean-valued
                # expression, if it turned out that the root of the
                # expression was just an element constraint. (This can
                # happen for something like a constraint that requires that
                # an interval var specified by indirection has to be
                # present.)
                cpx_model.add(expr[1] == True)
            else:
                cpx_model.add(expr[1])

        # That's all, folks.
        return cpx_model, visitor.pyomo_to_docplex


@SolverFactory.register('cp_optimizer', doc='Direct interface to CPLEX CP Optimizer')
class CPOptimizerSolver:
    CONFIG = ConfigDict("cp_optimizer_solver")
    CONFIG.declare(
        'symbolic_solver_labels',
        ConfigValue(
            default=False,
            domain=bool,
            description='Write Pyomo Var and Constraint names to docplex model',
        ),
    )
    CONFIG.declare(
        'tee',
        ConfigValue(
            default=False, domain=bool, description="Stream solver output to terminal."
        ),
    )
    CONFIG.declare(
        'options', ConfigValue(default={}, description="Dictionary of solver options.")
    )

    _unrestricted_license = None

    def __init__(self, **kwds):
        self.config = self.CONFIG()
        self.config.set_value(kwds)
        if docplex_available:
            self._solve_status_map = {
                cp.SOLVE_STATUS_UNKNOWN: TerminationCondition.unknown,
                cp.SOLVE_STATUS_INFEASIBLE: TerminationCondition.infeasible,
                cp.SOLVE_STATUS_FEASIBLE: TerminationCondition.feasible,
                cp.SOLVE_STATUS_OPTIMAL: TerminationCondition.optimal,
                cp.SOLVE_STATUS_JOB_ABORTED: None,  # we need the fail status
                cp.SOLVE_STATUS_JOB_FAILED: TerminationCondition.solverFailure,
            }
            self._stop_cause_map = {
                # We only need to check this if we get an 'aborted' status, so
                # if this says it hasn't been stopped, we're just confused at
                # this point.
                cp.STOP_CAUSE_NOT_STOPPED: TerminationCondition.unknown,
                cp.STOP_CAUSE_LIMIT: TerminationCondition.maxTimeLimit,
                # User called exit, maybe in a callback.
                cp.STOP_CAUSE_EXIT: TerminationCondition.userInterrupt,
                # docplex says "Search aborted externally"
                cp.STOP_CAUSE_ABORT: TerminationCondition.userInterrupt,
                # This is in their documentation, but not here, for some reason
                # cp.STOP_CAUSE_UNKNOWN: TerminationCondition.unknown
            }

    @property
    def options(self):
        return self.config.options

    # Support use as a context manager under current solver API
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    def available(self, exception_flag=True):
        return Executable('cpoptimizer').available() and docplex_available

    def license_is_valid(self):
        if CPOptimizerSolver._unrestricted_license is None:
            # Note: 140*log_2(140) == 998.1 fits in CE,
            #     141*log_2(141) == 1006.7 does not
            x = cp.integer_var_list(141, 1, 141, "X")
            m = cp.CpoModel()
            m.add(cp.all_diff(x))
            try:
                m.solve()
                CPOptimizerSolver._unrestricted_license = True
            except cp_solver.solver.CpoSolverException:
                CPOptimizerSolver._unrestricted_license = False
        return CPOptimizerSolver._unrestricted_license

    def solve(self, model, **kwds):
        """Solve the model.

        Args:
            model (Block): a Pyomo model or block to be solved

        """
        config = self.config()
        config.set_value(kwds)

        writer = DocplexWriter()
        cpx_model, var_map = writer.write(
            model, symbolic_solver_labels=config.symbolic_solver_labels
        )
        if not config.tee:
            # If the user has also set LogVerbosity, we'll assume they know what
            # they're doing.
            verbosity = config.options.get('LogVerbosity')
            if verbosity is None:
                config.options['LogVerbosity'] = 'Quiet'

        msol = cpx_model.solve(**self.options)

        # Transfer the solver status to the pyomo results object
        results = SolverResults()
        results.solver.name = "CP Optimizer"
        results.problem.name = model.name

        info = msol.get_solver_infos()
        results.problem.number_of_constraints = info.get_number_of_constraints()
        int_vars = info.get_number_of_integer_vars()
        interval_vars = info.get_number_of_interval_vars()
        results.problem.number_of_integer_vars = int_vars
        results.problem.number_of_interval_vars = interval_vars
        # This is a useless number, but so is 0, so...
        results.problem.number_of_variables = int_vars + interval_vars

        val = msol.get_objective_value()
        bound = msol.get_objective_bound()
        if cpx_model.is_maximization():
            results.problem.number_of_objectives = 1
            results.problem.sense = maximize
            results.problem.lower_bound = val
            results.problem.upper_bound = bound
        elif cpx_model.is_minimization():
            results.problem.number_of_objectives = 1
            results.problem.sense = minimize
            results.problem.lower_bound = bound
            results.problem.upper_bound = val
        else:
            # it's a satisfaction problem
            results.problem.number_of_objectives = 0
            results.problem.sense = None
            results.problem.lower_bound = None
            results.problem.upper_bound = None

        results.solver.solve_time = msol.get_solve_time()
        solve_status = msol.get_solve_status()
        results.solver.termination_condition = (
            self._solve_status_map[solve_status]
            if solve_status is not None
            else self._stop_cause_map[msol.get_stop_cause()]
        )

        # Copy the variable values onto the Pyomo model, using the map we stored
        # on the writer.
        cp_sol = msol.get_solution()
        if cp_sol is not None:
            for py_var, cp_var in var_map.items():
                sol = cp_sol.get_var_solution(cp_var)
                if sol is None:
                    logger.warning(
                        "CP optimizer did not return a value "
                        "for variable '%s'" % py_var.name
                    )
                else:
                    sol = sol.get_value()
                if py_var.ctype is SequenceVar:
                    # They don't actually have values--the IntervalVars will get
                    # set.
                    continue
                if py_var.ctype is IntervalVar:
                    if len(sol) == 0:
                        # The interval_var is absent
                        py_var.is_present.set_value(False)
                    else:
                        start, end, size = sol
                        py_var.is_present.set_value(True)
                        py_var.start_time.set_value(start, skip_validation=True)
                        py_var.end_time.set_value(end, skip_validation=True)
                        py_var.length.set_value(end - start, skip_validation=True)
                elif py_var.ctype in {Var, BooleanVar}:
                    py_var.set_value(sol, skip_validation=True)
                else:
                    raise DeveloperError(
                        "Unrecognized Pyomo type in pyomo-to-docplex "
                        "variable map: %s" % type(py_var)
                    )

        return results



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
This module defines the classes that provide an NLP interface based on
the Ampl Solver Library (ASL) implementation
"""

import os
import numpy as np

from scipy.sparse import coo_matrix
from pyomo.common.deprecation import deprecated
from pyomo.common.tempfiles import TempfileManager
from pyomo.opt import WriterFactory
import pyomo.core.base as pyo
from pyomo.common.collections import ComponentMap
from pyomo.common.env import CtypesEnviron
from pyomo.solvers.amplfunc_merge import amplfunc_merge
from ..sparse.block_matrix import BlockMatrix
from pyomo.contrib.pynumero.interfaces.ampl_nlp import AslNLP
from pyomo.contrib.pynumero.interfaces.nlp import NLP
from pyomo.core.base.suffix import SuffixFinder
from .external_grey_box import ExternalGreyBoxBlock


# TODO: There are todos in the code below
class PyomoNLP(AslNLP):
    def __init__(self, pyomo_model, nl_file_options=None):
        """
        Pyomo nonlinear program interface

        Parameters
        ----------
        pyomo_model: pyomo.environ.ConcreteModel
            Pyomo concrete model
        """
        TempfileManager.push()
        try:
            # get the temp file names for the nl file
            nl_file = TempfileManager.create_tempfile(suffix='pynumero.nl')

            # The current AmplInterface code only supports a single
            # objective function Therefore, we throw an error if there
            # is not one (and only one) active objective function. This
            # is better than adding a dummy objective that the user does
            # not know about (since we do not have a good place to
            # remove this objective later)
            #
            # TODO: extend the AmplInterface and the AslNLP to correctly
            # handle this
            #
            # This currently addresses issue #1217
            objectives = list(
                pyomo_model.component_data_objects(
                    ctype=pyo.Objective, active=True, descend_into=True
                )
            )
            if len(objectives) != 1:
                raise NotImplementedError(
                    'The ASL interface and PyomoNLP in PyNumero currently '
                    'only support single objective problems. Deactivate '
                    'any extra objectives you may have, or add a dummy '
                    'objective (f(x)=0) if you have a square problem '
                    '(found %s objectives).' % (len(objectives),)
                )
            self._objective = objectives[0]

            # write the nl file for the Pyomo model and get the symbolMap
            if nl_file_options is None:
                nl_file_options = dict()
            fname, symbolMap = WriterFactory('nl')(
                pyomo_model, nl_file, lambda x: True, nl_file_options
            )
            self._symbol_map = symbolMap

            # create component maps from vardata to idx and condata to idx
            self._vardata_to_idx = vdidx = ComponentMap()
            self._condata_to_idx = cdidx = ComponentMap()

            # TODO: Are these names totally consistent?
            for name, obj in symbolMap.bySymbol.items():
                if name[0] == 'v':
                    vdidx[obj] = int(name[1:])
                elif name[0] == 'c':
                    cdidx[obj] = int(name[1:])

            # The NL writer advertises the external function libraries
            # through the PYOMO_AMPLFUNC environment variable; merge it
            # with any preexisting AMPLFUNC definitions
            amplfunc = amplfunc_merge(os.environ)

            with CtypesEnviron(AMPLFUNC=amplfunc):
                super(PyomoNLP, self).__init__(nl_file)

            # keep pyomo model in cache
            self._pyomo_model = pyomo_model

            # Create ComponentMap corresponding to equality constraint indices
            # This must be done after the call to super-init.
            full_to_equality = self._con_full_eq_map
            equality_mask = self._con_full_eq_mask
            self._condata_to_eq_idx = ComponentMap(
                (con, full_to_equality[i])
                for con, i in self._condata_to_idx.items()
                if equality_mask[i]
            )
            full_to_inequality = self._con_full_ineq_map
            inequality_mask = self._con_full_ineq_mask
            self._condata_to_ineq_idx = ComponentMap(
                (con, full_to_inequality[i])
                for con, i in self._condata_to_idx.items()
                if inequality_mask[i]
            )

        finally:
            # delete the nl file
            TempfileManager.pop()

    @property
    def symbol_map(self):
        return self._symbol_map

    def pyomo_model(self):
        """
        Return optimization model
        """
        return self._pyomo_model

    def get_pyomo_objective(self):
        """
        Return an instance of the active objective function on the Pyomo model.
        (there can be only one)
        """
        return self._objective

    def get_pyomo_variables(self):
        """
        Return an ordered list of the Pyomo VarData objects in
        the order corresponding to the primals
        """
        # ToDo: is there a more efficient way to do this
        idx_to_vardata = {i: v for v, i in self._vardata_to_idx.items()}
        return [idx_to_vardata[i] for i in range(len(idx_to_vardata))]

    def get_pyomo_constraints(self):
        """
        Return an ordered list of the Pyomo ConData objects in
        the order corresponding to the primals
        """
        # ToDo: is there a more efficient way to do this
        idx_to_condata = {i: v for v, i in self._condata_to_idx.items()}
        return [idx_to_condata[i] for i in range(len(idx_to_condata))]

    def get_pyomo_equality_constraints(self):
        """
        Return an ordered list of the Pyomo ConData objects in
        the order corresponding to the equality constraints.
        """
        idx_to_condata = {i: c for c, i in self._condata_to_eq_idx.items()}
        return [idx_to_condata[i] for i in range(len(idx_to_condata))]

    def get_pyomo_inequality_constraints(self):
        """
        Return an ordered list of the Pyomo ConData objects in
        the order corresponding to the inequality constraints.
        """
        idx_to_condata = {i: c for c, i in self._condata_to_ineq_idx.items()}
        return [idx_to_condata[i] for i in range(len(idx_to_condata))]

    @deprecated(
        msg='This method has been replaced with primals_names',
        version='6.0.0',
        remove_in='6.0',
    )
    def variable_names(self):
        return self.primals_names()

    def primals_names(self):
        """
        Return an ordered list of the Pyomo variable
        names in the order corresponding to the primals
        """
        pyomo_variables = self.get_pyomo_variables()
        return [v.getname(fully_qualified=True) for v in pyomo_variables]

    def constraint_names(self):
        """
        Return an ordered list of the Pyomo constraint
        names in the order corresponding to internal constraint order
        """
        pyomo_constraints = self.get_pyomo_constraints()
        return [v.getname(fully_qualified=True) for v in pyomo_constraints]

    def equality_constraint_names(self):
        """
        Return an ordered list of the Pyomo ConData names in
        the order corresponding to the equality constraints.
        """
        equality_constraints = self.get_pyomo_equality_constraints()
        return [v.getname(fully_qualified=True) for v in equality_constraints]

    def inequality_constraint_names(self):
        """
        Return an ordered list of the Pyomo ConData names in
        the order corresponding to the inequality constraints.
        """
        inequality_constraints = self.get_pyomo_inequality_constraints()
        return [v.getname(fully_qualified=True) for v in inequality_constraints]

    def get_primal_indices(self, pyomo_variables):
        """
        Return the list of indices for the primals
        corresponding to the list of Pyomo variables provided

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        """
        assert isinstance(pyomo_variables, list)
        var_indices = []
        for v in pyomo_variables:
            if v.is_indexed():
                for vd in v.values():
                    var_id = self._vardata_to_idx[vd]
                    var_indices.append(var_id)
            else:
                var_id = self._vardata_to_idx[v]
                var_indices.append(var_id)
        return var_indices

    def get_constraint_indices(self, pyomo_constraints):
        """
        Return the list of indices for the constraints
        corresponding to the list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        assert isinstance(pyomo_constraints, list)
        con_indices = []
        for c in pyomo_constraints:
            if c.is_indexed():
                for cd in c.values():
                    con_id = self._condata_to_idx[cd]
                    con_indices.append(con_id)
            else:
                con_id = self._condata_to_idx[c]
                con_indices.append(con_id)
        return con_indices

    def get_equality_constraint_indices(self, constraints):
        """
        Return the list of equality indices for the constraints
        corresponding to the list of Pyomo constraints provided.

        Parameters
        ----------
        constraints : list of Pyomo Constraints or ConstraintData objects
        """
        indices = []
        for c in constraints:
            if c.is_indexed():
                for cd in c.values():
                    con_eq_idx = self._condata_to_eq_idx[cd]
                    indices.append(con_eq_idx)
            else:
                con_eq_idx = self._condata_to_eq_idx[c]
                indices.append(con_eq_idx)
        return indices

    def get_inequality_constraint_indices(self, constraints):
        """
        Return the list of inequality indices for the constraints
        corresponding to the list of Pyomo constraints provided.

        Parameters
        ----------
        constraints : list of Pyomo Constraints or ConstraintData objects
        """
        indices = []
        for c in constraints:
            if c.is_indexed():
                for cd in c.values():
                    con_ineq_idx = self._condata_to_ineq_idx[cd]
                    indices.append(con_ineq_idx)
            else:
                con_ineq_idx = self._condata_to_ineq_idx[c]
                indices.append(con_ineq_idx)
        return indices

    # overloaded from NLP
    def get_obj_scaling(self):
        scaling_finder = SuffixFinder(
            'scaling_factor', default=1.0, context=self._pyomo_model
        )
        val = scaling_finder.find(self.get_pyomo_objective())
        if not scaling_finder.all_suffixes:
            return None
        return val

    # overloaded from NLP
    def get_primals_scaling(self):
        scaling_finder = SuffixFinder(
            'scaling_factor', default=1.0, context=self._pyomo_model
        )
        primals_scaling = np.fromiter(
            (scaling_finder.find(v) for v in self.get_pyomo_variables()),
            count=self.n_primals(),
            dtype=float,
        )
        if not scaling_finder.all_suffixes:
            return None
        return primals_scaling

    # overloaded from NLP
    def get_constraints_scaling(self):
        scaling_finder = SuffixFinder(
            'scaling_factor', default=1.0, context=self._pyomo_model
        )
        constraints_scaling = np.fromiter(
            (scaling_finder.find(v) for v in self.get_pyomo_constraints()),
            count=self.n_constraints(),
            dtype=float,
        )
        if not scaling_finder.all_suffixes:
            return None
        return constraints_scaling

    def extract_subvector_grad_objective(self, pyomo_variables):
        """Compute the gradient of the objective and return the entries
        corresponding to the given Pyomo variables

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        """
        grad_obj = self.evaluate_grad_objective()
        return grad_obj[self.get_primal_indices(pyomo_variables)]

    def extract_subvector_constraints(self, pyomo_constraints):
        """
        Return the values of the constraints
        corresponding to the list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        residuals = self.evaluate_constraints()
        return residuals[self.get_constraint_indices(pyomo_constraints)]

    def extract_submatrix_jacobian(self, pyomo_variables, pyomo_constraints):
        """
        Return the submatrix of the jacobian that corresponds to the list
        of Pyomo variables and list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        jac = self.evaluate_jacobian()
        primal_indices = self.get_primal_indices(pyomo_variables)
        constraint_indices = self.get_constraint_indices(pyomo_constraints)
        row_mask = np.isin(jac.row, constraint_indices)
        col_mask = np.isin(jac.col, primal_indices)
        submatrix_mask = row_mask & col_mask
        submatrix_irows = np.compress(submatrix_mask, jac.row)
        submatrix_jcols = np.compress(submatrix_mask, jac.col)
        submatrix_data = np.compress(submatrix_mask, jac.data)

        # ToDo: this is expensive - have to think about how to do this with numpy
        row_submatrix_map = {j: i for i, j in enumerate(constraint_indices)}
        for i, v in enumerate(submatrix_irows):
            submatrix_irows[i] = row_submatrix_map[v]

        col_submatrix_map = {j: i for i, j in enumerate(primal_indices)}
        for i, v in enumerate(submatrix_jcols):
            submatrix_jcols[i] = col_submatrix_map[v]

        return coo_matrix(
            (submatrix_data, (submatrix_irows, submatrix_jcols)),
            shape=(len(constraint_indices), len(primal_indices)),
        )

    def extract_submatrix_hessian_lag(self, pyomo_variables_rows, pyomo_variables_cols):
        """
        Return the submatrix of the hessian of the lagrangian that
        corresponds to the list of Pyomo variables provided

        Parameters
        ----------
        pyomo_variables_rows : list of Pyomo Var or VarData objects
            List of Pyomo Var or VarData objects corresponding to the desired rows
        pyomo_variables_cols : list of Pyomo Var or VarData objects
            List of Pyomo Var or VarData objects corresponding to the desired columns
        """
        hess_lag = self.evaluate_hessian_lag()
        primal_indices_rows = self.get_primal_indices(pyomo_variables_rows)
        primal_indices_cols = self.get_primal_indices(pyomo_variables_cols)
        row_mask = np.isin(hess_lag.row, primal_indices_rows)
        col_mask = np.isin(hess_lag.col, primal_indices_cols)
        submatrix_mask = row_mask & col_mask
        submatrix_irows = np.compress(submatrix_mask, hess_lag.row)
        submatrix_jcols = np.compress(submatrix_mask, hess_lag.col)
        submatrix_data = np.compress(submatrix_mask, hess_lag.data)

        # ToDo: this is expensive - have to think about how to do this with numpy
        submatrix_map = {j: i for i, j in enumerate(primal_indices_rows)}
        for i, v in enumerate(submatrix_irows):
            submatrix_irows[i] = submatrix_map[v]

        submatrix_map = {j: i for i, j in enumerate(primal_indices_cols)}
        for i, v in enumerate(submatrix_jcols):
            submatrix_jcols[i] = submatrix_map[v]

        return coo_matrix(
            (submatrix_data, (submatrix_irows, submatrix_jcols)),
            shape=(len(primal_indices_rows), len(primal_indices_cols)),
        )

    def load_state_into_pyomo(self, bound_multipliers=None):
        primals = self.get_primals()
        variables = self.get_pyomo_variables()
        for var, val in zip(variables, primals):
            var.set_value(val)
        m = self.pyomo_model()
        model_suffixes = dict(pyo.suffix.active_import_suffix_generator(m))
        if 'dual' in model_suffixes:
            duals = self.get_duals()
            constraints = self.get_pyomo_constraints()
            model_suffixes['dual'].clear()
            model_suffixes['dual'].update(zip(constraints, duals))
        if 'ipopt_zL_out' in model_suffixes:
            model_suffixes['ipopt_zL_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zL_out'].update(
                    zip(variables, bound_multipliers[0])
                )
        if 'ipopt_zU_out' in model_suffixes:
            model_suffixes['ipopt_zU_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zU_out'].update(
                    zip(variables, bound_multipliers[1])
                )


# TODO: look for the [:-i] when i might be zero
class PyomoGreyBoxNLP(NLP):
    def __init__(self, pyomo_model):
        # store all the greybox custom block data objects
        greybox_components = []
        try:
            # We support Pynumero's ExternalGreyBoxBlock modeling
            # objects.  We need to find them and convert them to Blocks
            # before calling the NL writer so that the attached Vars get
            # picked up by the writer.
            for greybox in pyomo_model.component_objects(
                ExternalGreyBoxBlock, descend_into=True
            ):
                greybox.parent_block().reclassify_component_type(greybox, pyo.Block)
                greybox_components.append(greybox)

            self._pyomo_model = pyomo_model
            self._pyomo_nlp = PyomoNLP(pyomo_model)

        finally:
            # Restore the ctypes of the ExternalGreyBoxBlock components
            for greybox in greybox_components:
                greybox.parent_block().reclassify_component_type(
                    greybox, ExternalGreyBoxBlock
                )

        # get the greybox block data objects
        greybox_data = []
        for greybox in greybox_components:
            greybox_data.extend(data for data in greybox.values() if data.active)

        if len(greybox_data) > 1:
            raise NotImplementedError(
                "The PyomoGreyBoxModel interface has not"
                " been tested with Pyomo models that contain"
                " more than one ExternalGreyBoxBlock. Currently,"
                " only a single block is supported."
            )

        if self._pyomo_nlp.n_primals() == 0:
            raise ValueError(
                "No variables were found in the Pyomo part of the model."
                " PyomoGreyBoxModel requires at least one variable"
                " to be active in a Pyomo objective or constraint"
            )

        # check that the greybox model supports what we would expect
        # TODO: add support for models that do not provide jacobians
        """
        for data in greybox_data:
            c = data._ex_model.model_capabilities()
            if (c.n_equality_constraints() > 0 \
               and not c.supports_jacobian_equality_constraints) \
               or (c.n_equality_constraints() > 0 \
               and not c.supports_jacobian_equality_constraints)
                raise NotImplementedError('PyomoGreyBoxNLP does not support models'
                                          ' without explicit Jacobian support')
        """

        # number of additional variables required - they are in the
        # greybox models but not included in the NL file
        self._n_greybox_primals = 0

        # number of residuals (equality constraints + output constraints
        # coming from the grey box models
        self._greybox_primals_names = []
        self._greybox_constraints_names = []

        # Update the primal index map with any variables in the
        # greybox models that do not otherwise appear in the NL
        # and capture some other book keeping items
        n_primals = self._pyomo_nlp.n_primals()
        greybox_primals = []
        self._vardata_to_idx = ComponentMap(self._pyomo_nlp._vardata_to_idx)
        for data in greybox_data:
            # check that none of the inputs / outputs are fixed
            for v in data.inputs.values():
                if v.fixed:
                    raise NotImplementedError(
                        'Found a grey box model input that is fixed: {}.'
                        ' This interface does not currently support fixed'
                        ' variables. Please add a constraint instead.'
                        ''.format(v.getname(fully_qualified=True))
                    )
            for v in data.outputs.values():
                if v.fixed:
                    raise NotImplementedError(
                        'Found a grey box model output that is fixed: {}.'
                        ' This interface does not currently support fixed'
                        ' variables. Please add a constraint instead.'
                        ''.format(v.getname(fully_qualified=True))
                    )

            block_name = data.getname()
            for nm in data._ex_model.equality_constraint_names():
                self._greybox_constraints_names.append('{}.{}'.format(block_name, nm))
            for nm in data._ex_model.output_names():
                self._greybox_constraints_names.append(
                    '{}.{}_con'.format(block_name, nm)
                )

            for var in data.component_data_objects(pyo.Var):
                if var not in self._vardata_to_idx:
                    # there is a variable in the greybox block that
                    # is not in the NL - append this to the end
                    self._vardata_to_idx[var] = n_primals
                    n_primals += 1
                    greybox_primals.append(var)
                    self._greybox_primals_names.append(
                        var.getname(fully_qualified=True)
                    )
        self._n_greybox_primals = len(greybox_primals)
        self._greybox_primal_variables = greybox_primals

        # Configure the primal and dual data caches
        self._greybox_primals_lb = np.zeros(self._n_greybox_primals)
        self._greybox_primals_ub = np.zeros(self._n_greybox_primals)
        self._init_greybox_primals = np.zeros(self._n_greybox_primals)
        for i, var in enumerate(greybox_primals):
            if var.value is not None:
                self._init_greybox_primals[i] = var.value
            self._greybox_primals_lb[i] = -np.inf if var.lb is None else var.lb
            self._greybox_primals_ub[i] = np.inf if var.ub is None else var.ub
        self._greybox_primals_lb.flags.writeable = False
        self._greybox_primals_ub.flags.writeable = False

        self._greybox_primals = self._init_greybox_primals.copy()

        # data member to store the cached greybox constraints and jacobian
        self._cached_greybox_constraints = None
        self._cached_greybox_jac = None

        # create the helper objects
        con_offset = self._pyomo_nlp.n_constraints()
        self._external_greybox_helpers = []
        for data in greybox_data:
            h = _ExternalGreyBoxModelHelper(data, self._vardata_to_idx, con_offset)
            self._external_greybox_helpers.append(h)
            con_offset += h.n_residuals()

        self._n_greybox_constraints = con_offset - self._pyomo_nlp.n_constraints()
        assert len(self._greybox_constraints_names) == self._n_greybox_constraints

        # make sure the primal values get to the greybox models
        self.set_primals(self.get_primals())

        # If any part of the problem is scaled (i.e., obj, primals,
        # or any of the constraints for any of the external grey boxes),
        # then we want scaling factors for everything (defaulting to
        # ones for any of the missing factors).
        # This code builds all the scaling factors, with the defaults,
        # but then sets them all to None if *no* scaling factors are provided
        # for any of the pieces. (inefficient, but only done once)
        need_scaling = False
        self._obj_scaling = self._pyomo_nlp.get_obj_scaling()
        if self._obj_scaling is None:
            self._obj_scaling = 1.0
        else:
            need_scaling = True

        scaling_finder = SuffixFinder(
            'scaling_factor', default=1.0, context=self._pyomo_model
        )
        self._primals_scaling = np.fromiter(
            (scaling_finder.find(v) for v in self.get_pyomo_variables()),
            count=self.n_primals(),
            dtype=float,
        )
        need_scaling = bool(scaling_finder.all_suffixes)

        self._constraints_scaling = []
        pyomo_nlp_scaling = self._pyomo_nlp.get_constraints_scaling()
        if pyomo_nlp_scaling is None:
            pyomo_nlp_scaling = np.ones(self._pyomo_nlp.n_constraints())
        else:
            need_scaling = True
        self._constraints_scaling.append(pyomo_nlp_scaling)

        for h in self._external_greybox_helpers:
            tmp_scaling = h.get_residual_scaling()
            if tmp_scaling is None:
                tmp_scaling = np.ones(h.n_residuals())
            else:
                need_scaling = True
            self._constraints_scaling.append(tmp_scaling)

        if need_scaling:
            self._constraints_scaling = np.concatenate(self._constraints_scaling)
        else:
            self._obj_scaling = None
            self._primals_scaling = None
            self._constraints_scaling = None

        # might want the user to be able to specify these at some point
        self._init_greybox_duals = np.zeros(self._n_greybox_constraints)
        self._init_greybox_primals.flags.writeable = False
        self._init_greybox_duals.flags.writeable = False
        self._greybox_duals = self._init_greybox_duals.copy()

        # compute the jacobian for the external greybox models
        # to get some of the statistics
        self._evaluate_greybox_jacobians_and_cache_if_necessary()
        self._nnz_greybox_jac = len(self._cached_greybox_jac.data)

        # make sure the duals get to the external greybox models
        self.set_duals(self.get_duals())

        # compute the hessian for the external greybox models
        # to get some of the statistics
        try:
            self._evaluate_greybox_hessians_and_cache_if_necessary()
            self._nnz_greybox_hess = len(self._cached_greybox_hess.data)
        except (AttributeError, NotImplementedError):
            self._nnz_greybox_hess = None

    def _invalidate_greybox_primals_cache(self):
        self._greybox_constraints_cached = False
        self._greybox_jac_cached = False
        self._greybox_hess_cached = False

    def _invalidate_greybox_duals_cache(self):
        self._greybox_hess_cached = False

    # overloaded from NLP
    def n_primals(self):
        return self._pyomo_nlp.n_primals() + self._n_greybox_primals

    # overloaded from NLP
    def n_constraints(self):
        return self._pyomo_nlp.n_constraints() + self._n_greybox_constraints

    # overloaded from ExtendedNLP
    def n_eq_constraints(self):
        return self._pyomo_nlp.n_eq_constraints() + self._n_greybox_constraints

    # overloaded from ExtendedNLP
    def n_ineq_constraints(self):
        return self._pyomo_nlp.n_ineq_constraints()

    # overloaded from NLP
    def nnz_jacobian(self):
        return self._pyomo_nlp.nnz_jacobian() + self._nnz_greybox_jac

    # overloaded from AslNLP
    def nnz_jacobian_eq(self):
        return self._pyomo_nlp.nnz_jacobian_eq() + self._nnz_greybox_jac

    # overloaded from NLP
    def nnz_hessian_lag(self):
        return self._pyomo_nlp.nnz_hessian_lag() + self._nnz_greybox_hess

    # overloaded from NLP
    def primals_lb(self):
        return np.concatenate((self._pyomo_nlp.primals_lb(), self._greybox_primals_lb))

    # overloaded from NLP
    def primals_ub(self):
        return np.concatenate((self._pyomo_nlp.primals_ub(), self._greybox_primals_ub))

    # overloaded from NLP
    def constraints_lb(self):
        return np.concatenate(
            (
                self._pyomo_nlp.constraints_lb(),
                np.zeros(self._n_greybox_constraints, dtype=np.float64),
            )
        )

    # overloaded from NLP
    def constraints_ub(self):
        return np.concatenate(
            (
                self._pyomo_nlp.constraints_ub(),
                np.zeros(self._n_greybox_constraints, dtype=np.float64),
            )
        )

    # overloaded from NLP
    def init_primals(self):
        return np.concatenate(
            (self._pyomo_nlp.init_primals(), self._init_greybox_primals)
        )

    # overloaded from NLP
    def init_duals(self):
        return np.concatenate((self._pyomo_nlp.init_duals(), self._init_greybox_duals))

    # overloaded from ExtendedNLP
    def init_duals_eq(self):
        return np.concatenate(
            (self._pyomo_nlp.init_duals_eq(), self._init_greybox_duals)
        )

    # overloaded from NLP / Extended NLP
    def create_new_vector(self, vector_type):
        if vector_type == 'primals':
            return np.zeros(self.n_primals(), dtype=np.float64)
        elif vector_type == 'constraints' or vector_type == 'duals':
            return np.zeros(self.n_constraints(), dtype=np.float64)
        elif vector_type == 'eq_constraints' or vector_type == 'duals_eq':
            return np.zeros(self.n_eq_constraints(), dtype=np.float64)
        elif vector_type == 'ineq_constraints' or vector_type == 'duals_ineq':
            return np.zeros(self.n_ineq_constraints(), dtype=np.float64)
        else:
            raise RuntimeError('Called create_new_vector with an unknown vector_type')

    # overloaded from NLP
    def set_primals(self, primals):
        self._invalidate_greybox_primals_cache()

        # set the primals on the "pyomo" part of the nlp
        self._pyomo_nlp.set_primals(primals[: self._pyomo_nlp.n_primals()])

        # copy the values for the greybox primals
        np.copyto(self._greybox_primals, primals[self._pyomo_nlp.n_primals() :])

        for external in self._external_greybox_helpers:
            external.set_primals(primals)

    # overloaded from AslNLP
    def get_primals(self):
        # return the value of the primals that the pyomo
        # part knows about as well as any extra values that
        # are only in the greybox part
        return np.concatenate((self._pyomo_nlp.get_primals(), self._greybox_primals))

    # overloaded from NLP
    def set_duals(self, duals):
        self._invalidate_greybox_duals_cache()

        # set the duals for the pyomo part of the nlp
        self._pyomo_nlp.set_duals(duals[: self._pyomo_nlp.n_constraints()])

        # set the duals for the greybox part of the nlp
        np.copyto(self._greybox_duals, duals[self._pyomo_nlp.n_constraints() :])

        # set the duals in the helpers for the hessian computation
        for h in self._external_greybox_helpers:
            h.set_duals(duals)

    # overloaded from NLP
    def get_duals(self):
        # return the duals for the pyomo part of the nlp
        # concatenated with the greybox part
        return np.concatenate((self._pyomo_nlp.get_duals(), self._greybox_duals))

    # overloaded from ExtendedNLP
    def set_duals_eq(self, duals):
        raise NotImplementedError('set_duals_eq not implemented for PyomoGreyBoxNLP')
        # we think the code below is correct, but it has not yet been tested
        """
        #self._invalidate_greybox_duals_cache()

        # set the duals for the pyomo part of the nlp
        self._pyomo_nlp.set_duals_eq(
            duals[:self._pyomo_nlp.n_equality_constraints()])

        # set the duals for the greybox part of the nlp
        np.copyto(self._greybox_duals, duals[self._pyomo_nlp.n_equality_constraints():])
        # set the duals in the helpers for the hessian computation
        for h in self._external_greybox_helpers:
            h.set_duals_eq(duals)
        """

    # TODO: Implement set_duals_ineq

    # overloaded from NLP
    def get_duals_eq(self):
        raise NotImplementedError('get_duals_eq not implemented for PyomoGreyBoxNLP')
        """
        # return the duals for the pyomo part of the nlp
        # concatenated with the greybox part
        return np.concatenate((
            self._pyomo_nlp.get_duals_eq(),
            self._greybox_duals,
        ))
        """

    # overloaded from NLP
    def set_obj_factor(self, obj_factor):
        # objective is owned by the pyomo model
        self._pyomo_nlp.set_obj_factor(obj_factor)

    # overloaded from NLP
    def get_obj_factor(self):
        # objective is owned by the pyomo model
        return self._pyomo_nlp.get_obj_factor()

    # overloaded from NLP
    def get_obj_scaling(self):
        return self._obj_scaling

    # overloaded from NLP
    def get_primals_scaling(self):
        return self._primals_scaling

    # overloaded from NLP
    def get_constraints_scaling(self):
        return self._constraints_scaling

    # overloaded from NLP
    def evaluate_objective(self):
        # objective is owned by the pyomo model
        return self._pyomo_nlp.evaluate_objective()

    # overloaded from NLP
    def evaluate_grad_objective(self, out=None):
        # objective is owned by the pyomo model
        return np.concatenate(
            (
                self._pyomo_nlp.evaluate_grad_objective(out),
                np.zeros(self._n_greybox_primals),
            )
        )

    def _evaluate_greybox_constraints_and_cache_if_necessary(self):
        if self._greybox_constraints_cached:
            return

        self._cached_greybox_constraints = np.concatenate(
            tuple(
                external.evaluate_residuals()
                for external in self._external_greybox_helpers
            )
        )
        self._greybox_constraints_cached = True

    # overloaded from NLP
    def evaluate_constraints(self, out=None):
        self._evaluate_greybox_constraints_and_cache_if_necessary()

        if out is not None:
            if not isinstance(out, np.ndarray) or out.size != self.n_constraints():
                raise RuntimeError(
                    'Called evaluate_constraints with an invalid'
                    ' "out" argument - should take an ndarray of '
                    'size {}'.format(self.n_constraints())
                )

            # call on the pyomo part of the nlp
            self._pyomo_nlp.evaluate_constraints(out[: -self._n_greybox_constraints])

            # call on the greybox part of the nlp
            np.copyto(
                out[-self._n_greybox_constraints :], self._cached_greybox_constraints
            )
            return out

        else:
            # concatenate the pyomo and external constraint residuals
            return np.concatenate(
                (
                    self._pyomo_nlp.evaluate_constraints(),
                    self._cached_greybox_constraints,
                )
            )

    # overloaded from ExtendedNLP
    def evaluate_eq_constraints(self, out=None):
        raise NotImplementedError('Not yet implemented for PyomoGreyBoxNLP')
        """
        self._evaluate_greybox_constraints_and_cache_if_necessary()

        if out is not None:
            if not isinstance(out, np.ndarray) \
               or out.size != self.n_eq_constraints():
                raise RuntimeError(
                    'Called evaluate_eq_constraints with an invalid'
                    ' "out" argument - should take an ndarray of '
                    'size {}'.format(self.n_eq_constraints()))
            self._pyomo_nlp.evaluate_eq_constraints(
                out[:-self._n_greybox_constraints])
            np.copyto(out[-self._n_greybox_constraints:], self._cached_greybox_constraints)
            return out
        else:
            return np.concatenate((
                self._pyomo_nlp.evaluate_eq_constraints(),
                self._cached_greybox_constraints,
            ))
        """

    def _evaluate_greybox_jacobians_and_cache_if_necessary(self):
        if self._greybox_jac_cached:
            return

        jac = BlockMatrix(len(self._external_greybox_helpers), 1)
        for i, external in enumerate(self._external_greybox_helpers):
            jac.set_block(i, 0, external.evaluate_jacobian())
        self._cached_greybox_jac = jac.tocoo()
        self._greybox_jac_cached = True

    # overloaded from NLP
    def evaluate_jacobian(self, out=None):
        self._evaluate_greybox_jacobians_and_cache_if_necessary()

        if out is not None:
            if (
                not isinstance(out, coo_matrix)
                or out.shape[0] != self.n_constraints()
                or out.shape[1] != self.n_primals()
                or out.nnz != self.nnz_jacobian()
            ):
                raise RuntimeError(
                    'evaluate_jacobian called with an "out" argument'
                    ' that is invalid. This should be a coo_matrix with'
                    ' shape=({},{}) and nnz={}'.format(
                        self.n_constraints(), self.n_primals(), self.nnz_jacobian()
                    )
                )
            n_pyomo_constraints = self.n_constraints() - self._n_greybox_constraints

            # to avoid an additional copy, we pass in a slice (numpy view) of the underlying
            # data, row, and col that we were handed to be populated in evaluate_jacobian
            self._pyomo_nlp.evaluate_jacobian(
                out=coo_matrix(
                    (
                        out.data[: -self._nnz_greybox_jac],
                        (
                            out.row[: -self._nnz_greybox_jac],
                            out.col[: -self._nnz_greybox_jac],
                        ),
                    ),
                    shape=(n_pyomo_constraints, self._pyomo_nlp.n_primals()),
                )
            )
            np.copyto(out.data[-self._nnz_greybox_jac :], self._cached_greybox_jac.data)
            return out
        else:
            base = self._pyomo_nlp.evaluate_jacobian()
            base = coo_matrix(
                (base.data, (base.row, base.col)),
                shape=(base.shape[0], self.n_primals()),
            )

            jac = BlockMatrix(2, 1)
            jac.set_block(0, 0, base)
            jac.set_block(1, 0, self._cached_greybox_jac)
            return jac.tocoo()

            # TODO: Doesn't this need a "shape" specification?
            # return coo_matrix((
            #    np.concatenate((base.data, self._cached_greybox_jac.data)),
            #    ( np.concatenate((base.row, self._cached_greybox_jac.row)),
            #      np.concatenate((base.col, self._cached_greybox_jac.col)) )
            # ))

    # overloaded from ExtendedNLP
    """
    def evaluate_jacobian_eq(self, out=None):
        raise NotImplementedError()
        self._evaluate_greybox_jacobians_and_cache_if_necessary()

        if out is not None:
            if ( not isinstance(out, coo_matrix)
                 or out.shape[0] != self.n_eq_constraints()
                 or out.shape[1] != self.n_primals()
                 or out.nnz != self.nnz_jacobian_eq() ):
                raise RuntimeError(
                    'evaluate_jacobian called with an "out" argument'
                    ' that is invalid. This should be a coo_matrix with'
                    ' shape=({},{}) and nnz={}'
                    .format(self.n_eq_constraints(), self.n_primals(),
                            self.nnz_jacobian_eq()))
            self._pyomo_nlp.evaluate_jacobian_eq(
                coo_matrix((out.data[:-self._nnz_greybox_jac],
                            (out.row[:-self._nnz_greybox_jac],
                             out.col[:-self._nnz_greybox_jac])))
            )
            np.copyto(out.data[-self._nnz_greybox_jac],
                      self._cached_greybox_jac.data)
            return out
        else:
            base = self._pyomo_nlp.evaluate_jacobian_eq()
            # TODO: Doesn't this need a "shape" specification?
            return coo_matrix((
                np.concatenate((base.data, self._cached_greybox_jac.data)),
                ( np.concatenate((base.row, self._cached_greybox_jac.row)),
                  np.concatenate((base.col, self._cached_greybox_jac.col)) )
            ))
    """

    # overloaded from NLP
    def _evaluate_greybox_hessians_and_cache_if_necessary(self):
        if self._greybox_hess_cached:
            return

        data = list()
        irow = list()
        jcol = list()
        for external in self._external_greybox_helpers:
            hess = external.evaluate_hessian()
            data.append(hess.data)
            irow.append(hess.row)
            jcol.append(hess.col)

        data = np.concatenate(data)
        irow = np.concatenate(irow)
        jcol = np.concatenate(jcol)

        self._cached_greybox_hess = coo_matrix(
            (data, (irow, jcol)), shape=(self.n_primals(), self.n_primals())
        )
        self._greybox_hess_cached = True

    def evaluate_hessian_lag(self, out=None):
        self._evaluate_greybox_hessians_and_cache_if_necessary()
        if out is not None:
            if (
                not isinstance(out, coo_matrix)
                or out.shape[0] != self.n_primals()
                or out.shape[1] != self.n_primals()
                or out.nnz != self.nnz_hessian_lag()
            ):
                raise RuntimeError(
                    'evaluate_hessian_lag called with an "out" argument'
                    ' that is invalid. This should be a coo_matrix with'
                    ' shape=({},{}) and nnz={}'.format(
                        self.n_primals(), self.n_primals(), self.nnz_hessian()
                    )
                )
            # to avoid an additional copy, we pass in a slice (numpy view) of the underlying
            # data, row, and col that we were handed to be populated in evaluate_hessian_lag
            # the coo_matrix is simply a holder of the data, row, and col structures
            self._pyomo_nlp.evaluate_hessian_lag(
                out=coo_matrix(
                    (
                        out.data[: -self._nnz_greybox_hess],
                        (
                            out.row[: -self._nnz_greybox_hess],
                            out.col[: -self._nnz_greybox_hess],
                        ),
                    ),
                    shape=(self._pyomo_nlp.n_primals(), self._pyomo_nlp.n_primals()),
                )
            )
            np.copyto(
                out.data[-self._nnz_greybox_hess :], self._cached_greybox_hess.data
            )
            return out
        else:
            hess = self._pyomo_nlp.evaluate_hessian_lag()
            data = np.concatenate((hess.data, self._cached_greybox_hess.data))
            row = np.concatenate((hess.row, self._cached_greybox_hess.row))
            col = np.concatenate((hess.col, self._cached_greybox_hess.col))
            hess = coo_matrix(
                (data, (row, col)), shape=(self.n_primals(), self.n_primals())
            )
            return hess

    # overloaded from NLP
    def report_solver_status(self, status_code, status_message):
        raise NotImplementedError('Todo: implement this')

    @deprecated(
        msg='This method has been replaced with primals_names',
        version='6.0.0',
        remove_in='6.0',
    )
    def variable_names(self):
        return self.primals_names()

    def primals_names(self):
        names = list(self._pyomo_nlp.variable_names())
        names.extend(self._greybox_primals_names)
        return names

    def constraint_names(self):
        names = list(self._pyomo_nlp.constraint_names())
        names.extend(self._greybox_constraints_names)
        return names

    def pyomo_model(self):
        """
        Return optimization model
        """
        return self._pyomo_model

    def get_pyomo_objective(self):
        """
        Return an instance of the active objective function on the Pyomo model.
        (there can be only one)
        """
        return self._pyomo_nlp.get_pyomo_objective()

    def get_pyomo_variables(self):
        """
        Return an ordered list of the Pyomo VarData objects in
        the order corresponding to the primals
        """
        return self._pyomo_nlp.get_pyomo_variables() + self._greybox_primal_variables

    def get_pyomo_constraints(self):
        """
        Return an ordered list of the Pyomo ConData objects in
        the order corresponding to the primals
        """
        # FIXME: what do we return for the external block constraints?
        # return self._pyomo_nlp.get_pyomo_constraints()
        raise NotImplementedError(
            "returning list of all constraints when using an external model is TBD"
        )

    def load_state_into_pyomo(self, bound_multipliers=None):
        primals = self.get_primals()
        variables = self.get_pyomo_variables()
        for var, val in zip(variables, primals):
            var.set_value(val)
        m = self.pyomo_model()
        model_suffixes = dict(pyo.suffix.active_import_suffix_generator(m))
        if 'dual' in model_suffixes:
            model_suffixes['dual'].clear()
            # Until we sort out how to return the duals for the external
            # block (implied) constraints, I am disabling *all* duals
            #
            # duals = self.get_duals()
            # constraints = self.get_pyomo_constraints()
            # model_suffixes['dual'].update(
            #     zip(constraints, duals))
        if 'ipopt_zL_out' in model_suffixes:
            model_suffixes['ipopt_zL_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zL_out'].update(
                    zip(variables, bound_multipliers[0])
                )
        if 'ipopt_zU_out' in model_suffixes:
            model_suffixes['ipopt_zU_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zU_out'].update(
                    zip(variables, bound_multipliers[1])
                )


class _ExternalGreyBoxModelHelper:
    def __init__(self, ex_grey_box_block, vardata_to_idx, con_offset):
        """This helper takes an ExternalGreyBoxModel and provides the residual,
        Jacobian, and Hessian computation mapped to the correct variable space.

        The ExternalGreyBoxModel provides an interface that supports
        equality constraints (pure residuals) and output equations. Let
        u be the inputs, o be the outputs, and x be the full set of
        primal variables from the entire pyomo_nlp.

        With this, the ExternalGreyBoxModel provides the residual
        computations w_eq(u), and w_o(u), as well as the Jacobians,
        Jw_eq(u), and Jw_o(u). This helper provides h(x)=0, where h(x) =
        [h_eq(x); h_o(x)-o] and h_eq(x)=w_eq(Pu*x), and
        h_o(x)=w_o(Pu*x), and Pu is a mapping from the full primal
        variables "x" to the inputs "u".

        It also provides the Jacobian of h w.r.t. x.
           J_h(x) = [Jw_eq(Pu*x); Jw_o(Pu*x)-Po*x]
        where Po is a mapping from the full primal variables "x" to the
        outputs "o".

        """
        self._block = ex_grey_box_block
        self._ex_model = ex_grey_box_block.get_external_model()
        self._n_primals = len(vardata_to_idx)
        n_inputs = len(self._block.inputs)
        assert n_inputs == self._ex_model.n_inputs()
        n_eq_constraints = self._ex_model.n_equality_constraints()
        n_outputs = len(self._block.outputs)
        assert n_outputs == self._ex_model.n_outputs()

        # store the map of input indices (0 .. n_inputs) to
        # the indices in the full primals vector
        self._inputs_to_primals_map = np.fromiter(
            (vardata_to_idx[v] for v in self._block.inputs.values()),
            dtype=np.int64,
            count=n_inputs,
        )

        # store the map of output indices (0 .. n_outputs) to
        # the indices in the full primals vector
        self._outputs_to_primals_map = np.fromiter(
            (vardata_to_idx[v] for v in self._block.outputs.values()),
            dtype=np.int64,
            count=n_outputs,
        )

        if (
            self._ex_model.n_outputs() == 0
            and self._ex_model.n_equality_constraints() == 0
        ):
            raise ValueError(
                'ExternalGreyBoxModel has no equality constraints '
                'or outputs. It must have at least one or both.'
            )

        self._ex_eq_duals_to_full_map = None
        if n_eq_constraints > 0:
            self._ex_eq_duals_to_full_map = list(
                range(con_offset, con_offset + n_eq_constraints)
            )

        self._ex_output_duals_to_full_map = None
        if n_outputs > 0:
            self._ex_output_duals_to_full_map = list(
                range(
                    con_offset + n_eq_constraints,
                    con_offset + n_eq_constraints + n_outputs,
                )
            )

        # we need to change the column indices in the jacobian
        # from the 0..n_inputs provided by the external model
        # to the indices corresponding to the full Pyomo model
        # so we create that here
        self._eq_jac_primal_jcol = None
        self._outputs_jac_primal_jcol = None
        self._additional_output_entries_irow = None
        self._additional_output_entries_jcol = None
        self._additional_output_entries_data = None
        self._con_offset = con_offset
        self._eq_hess_jcol = None
        self._eq_hess_irow = None
        self._output_hess_jcol = None
        self._output_hess_irow = None

    def set_primals(self, primals):
        # map the full primals "x" to the inputs "u" and set
        # the values on the external model
        input_values = primals[self._inputs_to_primals_map]
        self._ex_model.set_input_values(input_values)

        # map the full primals "x" to the outputs "o" and
        # store a vector of the current output values to
        # use when evaluating residuals
        self._output_values = primals[self._outputs_to_primals_map]

    def set_duals(self, duals):
        # map the full duals to the duals for the equality
        # and the output constraints
        if self._ex_eq_duals_to_full_map is not None:
            duals_eq = duals[self._ex_eq_duals_to_full_map]
            self._ex_model.set_equality_constraint_multipliers(duals_eq)

        if self._ex_output_duals_to_full_map is not None:
            duals_outputs = duals[self._ex_output_duals_to_full_map]
            self._ex_model.set_output_constraint_multipliers(duals_outputs)

    def n_residuals(self):
        return self._ex_model.n_equality_constraints() + self._ex_model.n_outputs()

    def get_residual_scaling(self):
        eq_scaling = self._ex_model.get_equality_constraint_scaling_factors()
        output_con_scaling = self._ex_model.get_output_constraint_scaling_factors()
        if eq_scaling is None and output_con_scaling is None:
            return None
        if eq_scaling is None:
            eq_scaling = np.ones(self._ex_model.n_equality_constraints())
        if output_con_scaling is None:
            output_con_scaling = np.ones(self._ex_model.n_outputs())

        return np.concatenate((eq_scaling, output_con_scaling))

    def evaluate_residuals(self):
        # evaluate the equality constraints and the output equations
        # and return a single vector of residuals
        # returns residual for h(x)=0, where h(x) = [h_eq(x); h_o(x)-o]
        resid_list = []
        if self._ex_model.n_equality_constraints() > 0:
            resid_list.append(self._ex_model.evaluate_equality_constraints())

        if self._ex_model.n_outputs() > 0:
            computed_output_values = self._ex_model.evaluate_outputs()
            output_resid = computed_output_values - self._output_values
            resid_list.append(output_resid)

        return np.concatenate(resid_list)

    def evaluate_jacobian(self):
        # compute the jacobian of h(x) w.r.t. x
        # J_h(x) = [Jw_eq(Pu*x); Jw_o(Pu*x)-Po*x]

        # Jw_eq(x)
        eq_jac = None
        if self._ex_model.n_equality_constraints() > 0:
            eq_jac = self._ex_model.evaluate_jacobian_equality_constraints()
            if self._eq_jac_primal_jcol is None:
                # The first time through, we won't have created the
                # mapping of external primals ('u') to the full space
                # primals ('x')
                self._eq_jac_primal_jcol = self._inputs_to_primals_map[eq_jac.col]
            # map the columns from the inputs "u" back to the full primals "x"
            eq_jac = coo_matrix(
                (eq_jac.data, (eq_jac.row, self._eq_jac_primal_jcol)),
                (eq_jac.shape[0], self._n_primals),
            )

        outputs_jac = None
        if self._ex_model.n_outputs() > 0:
            outputs_jac = self._ex_model.evaluate_jacobian_outputs()

            row = outputs_jac.row
            # map the columns from the inputs "u" back to the full primals "x"
            if self._outputs_jac_primal_jcol is None:
                # The first time through, we won't have created the
                # mapping of external outputs ('o') to the full space
                # primals ('x')
                self._outputs_jac_primal_jcol = self._inputs_to_primals_map[
                    outputs_jac.col
                ]

                # We also need tocreate the irow, jcol, nnz structure for the
                # output variable portion of h(u)-o=0
                self._additional_output_entries_irow = np.asarray(
                    range(self._ex_model.n_outputs())
                )
                self._additional_output_entries_jcol = self._outputs_to_primals_map
                self._additional_output_entries_data = -1.0 * np.ones(
                    self._ex_model.n_outputs()
                )

            col = self._outputs_jac_primal_jcol
            data = outputs_jac.data

            # add the additional entries for the -Po*x portion of the jacobian
            row = np.concatenate((row, self._additional_output_entries_irow))
            col = np.concatenate((col, self._additional_output_entries_jcol))
            data = np.concatenate((data, self._additional_output_entries_data))
            outputs_jac = coo_matrix(
                (data, (row, col)), shape=(outputs_jac.shape[0], self._n_primals)
            )

        jac = None
        if eq_jac is not None:
            if outputs_jac is not None:
                # create a jacobian with both Jw_eq and Jw_o
                jac = BlockMatrix(2, 1)
                jac.name = 'external model jacobian'
                jac.set_block(0, 0, eq_jac)
                jac.set_block(1, 0, outputs_jac)
            else:
                assert self._ex_model.n_outputs() == 0
                assert self._ex_model.n_equality_constraints() > 0
                # only need the Jacobian with Jw_eq (there are not
                # output equations)
                jac = eq_jac
        else:
            assert outputs_jac is not None
            assert self._ex_model.n_outputs() > 0
            assert self._ex_model.n_equality_constraints() == 0
            # only need the Jacobian with Jw_o (there are no equalities)
            jac = outputs_jac

        return jac

    def evaluate_hessian(self):
        # compute the portion of the Hessian of the Lagrangian
        # of h(x) w.r.t. x
        # H_h(x) = sum_i y_eq_i * Hw_eq(Pu*x) +_ sum_k y_o_j * Hw_o(Pu*x)]

        data_list = []
        irow_list = []
        jcol_list = []

        # Hw_eq(x)
        eq_hess = None
        if self._ex_model.n_equality_constraints() > 0:
            eq_hess = self._ex_model.evaluate_hessian_equality_constraints()
            if self._eq_hess_jcol is None:
                # first time through, let's also check that it is lower triangular
                if np.any(eq_hess.row < eq_hess.col):
                    raise ValueError(
                        'ExternalGreyBoxModel must return lower '
                        'triangular portion of the Hessian only'
                    )

                # The first time through, we won't have created the
                # mapping of external primals ('u') to the full space
                # primals ('x')
                self._eq_hess_irow = row = self._inputs_to_primals_map[eq_hess.row]
                self._eq_hess_jcol = col = self._inputs_to_primals_map[eq_hess.col]

                # mapping may have made this not lower triangular
                mask = col > row
                row[mask], col[mask] = col[mask], row[mask]

            data_list.append(eq_hess.data)
            irow_list.append(self._eq_hess_irow)
            jcol_list.append(self._eq_hess_jcol)

        if self._ex_model.n_outputs() > 0:
            output_hess = self._ex_model.evaluate_hessian_outputs()
            if self._output_hess_irow is None:
                # first time through, let's also check that it is lower triangular
                if np.any(output_hess.row < output_hess.col):
                    raise ValueError(
                        'ExternalGreyBoxModel must return lower '
                        'triangular portion of the Hessian only'
                    )

                # The first time through, we won't have created the
                # mapping of external outputs ('o') to the full space
                # primals ('x')
                self._output_hess_irow = row = self._inputs_to_primals_map[
                    output_hess.row
                ]
                self._output_hess_jcol = col = self._inputs_to_primals_map[
                    output_hess.col
                ]

                # mapping may have made this not lower triangular
                mask = col > row
                row[mask], col[mask] = col[mask], row[mask]

            data_list.append(output_hess.data)
            irow_list.append(self._output_hess_irow)
            jcol_list.append(self._output_hess_jcol)

        data = np.concatenate(data_list)
        irow = np.concatenate(irow_list)
        jcol = np.concatenate(jcol_list)
        hess = coo_matrix((data, (irow, jcol)), (self._n_primals, self._n_primals))

        return hess



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo.common.unittest as unittest

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as sp

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run BlockMatrix tests")

from scipy.sparse import coo_matrix, bmat

from pyomo.contrib.pynumero.sparse import (
    BlockMatrix,
    BlockVector,
    NotFullyDefinedBlockMatrixError,
)
import warnings


class TestBlockMatrix(unittest.TestCase):
    def setUp(self):
        row = np.array([0, 3, 1, 2, 3, 0])
        col = np.array([0, 0, 1, 2, 3, 3])
        data = np.array([2.0, 1, 3, 4, 5, 1])
        m = coo_matrix((data, (row, col)), shape=(4, 4))

        self.block_m = m

        bm = BlockMatrix(2, 2)
        bm.name = 'basic_matrix'
        bm.set_block(0, 0, m.copy())
        bm.set_block(1, 1, m.copy())
        bm.set_block(0, 1, m.copy())
        self.basic_m = bm
        self.dense = np.zeros((8, 8))
        self.dense[0:4, 0:4] = m.toarray()
        self.dense[0:4, 4:8] = m.toarray()
        self.dense[4:8, 4:8] = m.toarray()

        self.composed_m = BlockMatrix(2, 2)
        self.composed_m.set_block(0, 0, self.block_m.copy())
        self.composed_m.set_block(1, 1, self.basic_m.copy())

    def test_name(self):
        self.assertEqual(self.basic_m.name, 'basic_matrix')
        self.basic_m.name = 'hola'
        self.assertEqual(self.basic_m.name, 'hola')

    def test_bshape(self):
        self.assertEqual(self.basic_m.bshape, (2, 2))

    def test_shape(self):
        shape = (self.block_m.shape[0] * 2, self.block_m.shape[1] * 2)
        self.assertEqual(self.basic_m.shape, shape)

    def test_tocoo(self):
        block = self.block_m
        m = self.basic_m
        scipy_mat = bmat([[block, block], [None, block]], format='coo')
        dinopy_mat = m.tocoo()
        drow = np.sort(dinopy_mat.row)
        dcol = np.sort(dinopy_mat.col)
        ddata = np.sort(dinopy_mat.data)
        srow = np.sort(scipy_mat.row)
        scol = np.sort(scipy_mat.col)
        sdata = np.sort(scipy_mat.data)
        self.assertListEqual(drow.tolist(), srow.tolist())
        self.assertListEqual(dcol.tolist(), scol.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

    def test_tocsr(self):
        block = self.block_m
        m = self.basic_m
        scipy_mat = bmat([[block, block], [None, block]], format='csr')
        dinopy_mat = m.tocsr()
        dindices = np.sort(dinopy_mat.indices)
        dindptr = np.sort(dinopy_mat.indptr)
        ddata = np.sort(dinopy_mat.data)
        sindices = np.sort(scipy_mat.indices)
        sindptr = np.sort(scipy_mat.indptr)
        sdata = np.sort(scipy_mat.data)
        self.assertListEqual(dindices.tolist(), sindices.tolist())
        self.assertListEqual(dindptr.tolist(), sindptr.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

    def test_tocsc(self):
        block = self.block_m
        m = self.basic_m
        scipy_mat = bmat([[block, block], [None, block]], format='csc')
        dinopy_mat = m.tocsc()
        dindices = np.sort(dinopy_mat.indices)
        dindptr = np.sort(dinopy_mat.indptr)
        ddata = np.sort(dinopy_mat.data)
        sindices = np.sort(scipy_mat.indices)
        sindptr = np.sort(scipy_mat.indptr)
        sdata = np.sort(scipy_mat.data)
        self.assertListEqual(dindices.tolist(), sindices.tolist())
        self.assertListEqual(dindptr.tolist(), sindptr.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

    def test_multiply(self):
        # check scalar multiplication
        block = self.block_m
        m = self.basic_m * 5.0
        scipy_mat = bmat([[block, block], [None, block]], format='coo')
        mulscipy_mat = scipy_mat * 5.0
        dinopy_mat = m.tocoo()
        drow = np.sort(dinopy_mat.row)
        dcol = np.sort(dinopy_mat.col)
        ddata = np.sort(dinopy_mat.data)
        srow = np.sort(mulscipy_mat.row)
        scol = np.sort(mulscipy_mat.col)
        sdata = np.sort(mulscipy_mat.data)
        self.assertListEqual(drow.tolist(), srow.tolist())
        self.assertListEqual(dcol.tolist(), scol.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

        m = 5.0 * self.basic_m
        dinopy_mat = m.tocoo()
        drow = np.sort(dinopy_mat.row)
        dcol = np.sort(dinopy_mat.col)
        ddata = np.sort(dinopy_mat.data)
        self.assertListEqual(drow.tolist(), srow.tolist())
        self.assertListEqual(dcol.tolist(), scol.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

        # check dot product with block vector
        block = self.block_m
        m = self.basic_m
        scipy_mat = bmat([[block, block], [None, block]], format='coo')
        x = BlockVector(2)
        x.set_block(0, np.ones(block.shape[1], dtype=np.float64))
        x.set_block(1, np.ones(block.shape[1], dtype=np.float64))

        res_scipy = scipy_mat.dot(x.flatten())
        res_dinopy = m * x
        res_dinopy_flat = m * x.flatten()

        self.assertListEqual(res_dinopy.tolist(), res_scipy.tolist())
        self.assertListEqual(res_dinopy_flat.tolist(), res_scipy.tolist())

        dense_mat = dinopy_mat.toarray()
        self.basic_m *= 5.0
        self.assertTrue(np.allclose(dense_mat, self.basic_m.toarray()))

    def test_mul_sparse_matrix(self):
        m = self.basic_m

        flat_prod = m.tocoo() * m.tocoo()
        prod = m * m

        self.assertIsInstance(prod, BlockMatrix)
        self.assertTrue(np.allclose(flat_prod.toarray(), prod.toarray()))

        m2 = m.copy_structure()
        ones = np.ones(m.shape)
        m2.copyfrom(ones)
        flat_prod = m.tocoo() * m2.tocoo()
        prod = m * m2

        self.assertIsInstance(prod, BlockMatrix)
        self.assertTrue(np.allclose(flat_prod.toarray(), prod.toarray()))

    def test_getitem(self):
        m = BlockMatrix(3, 3)
        for i in range(3):
            for j in range(3):
                self.assertIsNone(m.get_block(i, j))

        m.set_block(0, 1, self.block_m)
        self.assertEqual(m.get_block(0, 1).shape, self.block_m.shape)

    def test_setitem(self):
        m = BlockMatrix(2, 2)
        m.set_block(0, 1, self.block_m)
        self.assertFalse(m.is_empty_block(0, 1))
        self.assertEqual(m._brow_lengths[0], self.block_m.shape[0])
        self.assertEqual(m._bcol_lengths[1], self.block_m.shape[1])
        self.assertEqual(m.get_block(0, 1).shape, self.block_m.shape)

    def test_coo_data(self):
        m = self.basic_m.tocoo()
        data = self.basic_m.coo_data()
        self.assertListEqual(m.data.tolist(), data.tolist())

    # ToDo: add tests for block matrices with block matrices in it
    # ToDo: add tests for matrices with zeros in the diagonal
    # ToDo: add tests for block matrices with coo and csc matrices

    def test_nnz(self):
        self.assertEqual(self.block_m.nnz * 3, self.basic_m.nnz)

    def test_block_shapes(self):
        shapes = self.basic_m.block_shapes()
        for i in range(self.basic_m.bshape[0]):
            for j in range(self.basic_m.bshape[1]):
                self.assertEqual(shapes[i][j], self.block_m.shape)

    def test_dot(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        x = np.ones(A_dense.shape[1])
        block_x = BlockVector(2)
        block_x.set_block(0, np.ones(self.block_m.shape[1]))
        block_x.set_block(1, np.ones(self.block_m.shape[1]))
        flat_res = A_block.dot(x).flatten()
        block_res = A_block.dot(block_x)
        self.assertTrue(np.allclose(A_dense.dot(x), flat_res))
        self.assertTrue(np.allclose(A_dense.dot(x), block_res.flatten()))
        self.assertEqual(block_res.bshape[0], 2)

        m = BlockMatrix(2, 2)
        sub_m = np.array([[1, 0], [0, 1]])
        sub_m = coo_matrix(sub_m)
        m.set_block(0, 1, sub_m.copy())
        m.set_block(1, 0, sub_m.copy())
        x = np.arange(4)
        res = m * x
        self.assertTrue(np.allclose(res.flatten(), np.array([2, 3, 0, 1])))

    def test_reset_brow(self):
        self.basic_m.reset_brow(0)
        for j in range(self.basic_m.bshape[1]):
            self.assertIsNone(self.basic_m.get_block(0, j))

    def test_reset_bcol(self):
        self.basic_m.reset_bcol(0)
        for j in range(self.basic_m.bshape[0]):
            self.assertIsNone(self.basic_m.get_block(j, 0))

    def test_to_scipy(self):
        block = self.block_m
        m = self.basic_m
        scipy_mat = bmat([[block, block], [None, block]], format='coo')
        dinopy_mat = m.tocoo()
        drow = np.sort(dinopy_mat.row)
        dcol = np.sort(dinopy_mat.col)
        ddata = np.sort(dinopy_mat.data)
        srow = np.sort(scipy_mat.row)
        scol = np.sort(scipy_mat.col)
        sdata = np.sort(scipy_mat.data)
        self.assertListEqual(drow.tolist(), srow.tolist())
        self.assertListEqual(dcol.tolist(), scol.tolist())
        self.assertListEqual(ddata.tolist(), sdata.tolist())

    def test_has_undefined_row_sizes(self):
        self.assertFalse(self.basic_m.has_undefined_row_sizes())

    def test_has_undefined_col_sizes(self):
        self.assertFalse(self.basic_m.has_undefined_col_sizes())

    def test_transpose(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        A_dense_t = A_dense.transpose()
        A_block_t = A_block.transpose()
        self.assertTrue(np.allclose(A_dense_t, A_block_t.toarray()))

        A_dense = self.composed_m.toarray()
        A_block = self.composed_m
        A_dense_t = A_dense.transpose()
        A_block_t = A_block.transpose()
        self.assertTrue(np.allclose(A_dense_t, A_block_t.toarray()))

    def test_repr(self):
        self.assertEqual(len(self.basic_m.__repr__()), 17)

    def test_set_item(self):
        self.basic_m.set_block(1, 0, None)
        self.assertIsNone(self.basic_m.get_block(1, 0))
        self.basic_m.set_block(1, 1, None)
        self.assertIsNone(self.basic_m.get_block(1, 1))
        self.assertEqual(self.basic_m._brow_lengths[1], self.block_m.shape[0])
        self.basic_m.set_block(1, 1, self.block_m)
        self.assertEqual(self.basic_m._brow_lengths[1], self.block_m.shape[0])

    def test_add(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m

        aa = A_dense + A_dense
        mm = A_block + A_block

        self.assertTrue(np.allclose(aa, mm.toarray()))

        mm = A_block.__radd__(A_block)
        self.assertTrue(np.allclose(aa, mm.toarray()))

        r = A_block + A_block.tocoo()
        dense_res = A_block.toarray() + A_block.toarray()
        self.assertIsInstance(r, BlockMatrix)
        self.assertTrue(np.allclose(r.toarray(), dense_res))

        r = A_block.tocoo() + A_block
        dense_res = A_block.toarray() + A_block.toarray()
        # self.assertIsInstance(r, BlockMatrix)
        self.assertTrue(np.allclose(r.toarray(), dense_res))

        r = A_block + 2 * A_block.tocoo()
        dense_res = A_block.toarray() + 2 * A_block.toarray()
        self.assertIsInstance(r, BlockMatrix)
        self.assertTrue(np.allclose(r.toarray(), dense_res))

        r = 2 * A_block.tocoo() + A_block
        dense_res = 2 * A_block.toarray() + A_block.toarray()
        # self.assertIsInstance(r, BlockMatrix)
        self.assertTrue(np.allclose(r.toarray(), dense_res))

        r = A_block.T + A_block.tocoo()
        dense_res = A_block.toarray().T + A_block.toarray()
        self.assertIsInstance(r, BlockMatrix)
        self.assertTrue(np.allclose(r.toarray(), dense_res))

        with self.assertRaises(Exception) as context:
            mm = A_block.toarray() + A_block

        with self.assertRaises(Exception) as context:
            mm = A_block + A_block.toarray()

        with self.assertRaises(Exception) as context:
            mm = A_block + 1.0

    def test_add_copy(self):
        """
        The purpose of this test is to ensure that copying happens correctly when block matrices are added.
        For example, when adding

        [A  B   +  [D  0
         0  C]      E  F]

        we want to make sure that E and B both get copied in the result rather than just placed in the result.
        """
        bm = self.basic_m.copy()
        bmT = bm.transpose()
        res = bm + bmT
        self.assertIsNot(res.get_block(1, 0), bmT.get_block(1, 0))
        self.assertIsNot(res.get_block(0, 1), bm.get_block(0, 1))
        self.assertTrue(np.allclose(res.toarray(), self.dense + self.dense.transpose()))

    def test_sub(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        A_block2 = 2 * self.basic_m

        aa = A_dense - A_dense
        mm = A_block - A_block

        self.assertTrue(np.allclose(aa, mm.toarray()))

        mm = A_block2 - A_block.tocoo()
        self.assertTrue(np.allclose(A_block.toarray(), mm.toarray()))

        mm = A_block2.tocoo() - A_block
        self.assertTrue(np.allclose(A_block.toarray(), mm.toarray()))

        mm = A_block2.T - A_block.tocoo()
        dense_r = A_block2.toarray().T - A_block.toarray()
        self.assertTrue(np.allclose(dense_r, mm.toarray()))

        with self.assertRaises(Exception) as context:
            mm = A_block - A_block.toarray()

        with self.assertRaises(Exception) as context:
            mm = A_block - 1.0

        with self.assertRaises(Exception) as context:
            mm = 1.0 - A_block

    def test_sub_copy(self):
        """
        The purpose of this test is to ensure that copying happens correctly when block matrices are subtracted.
        For example, when subtracting

        [A  B   -  [D  0
         0  C]      E  F]

        we want to make sure that E and B both get copied in the result rather than just placed in the result.
        """
        bm = self.basic_m.copy()
        bmT = 2 * bm.transpose()
        res = bm - bmT
        self.assertIsNot(res.get_block(1, 0), bmT.get_block(1, 0))
        self.assertIsNot(res.get_block(0, 1), bm.get_block(0, 1))
        self.assertTrue(
            np.allclose(res.toarray(), self.dense - 2 * self.dense.transpose())
        )

    def test_neg(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m

        aa = -A_dense
        mm = -A_block

        self.assertTrue(np.allclose(aa, mm.toarray()))

    def test_copyfrom(self):
        bm0 = self.basic_m.copy()
        bm = bm0.copy_structure()
        self.assertFalse(np.allclose(bm.toarray(), self.dense))
        bm.copyfrom(bm0.tocoo())
        self.assertTrue(np.allclose(bm.toarray(), self.dense))

        flat = np.ones((8, 8))
        bm.copyfrom(flat)
        self.assertTrue(np.allclose(flat, bm.toarray()))

        bm.copyfrom(bm0)
        self.assertTrue(np.allclose(bm.toarray(), self.dense))

        bm.get_block(0, 0).data.fill(1.0)
        self.assertAlmostEqual(
            bm0.toarray()[0, 0], 2
        )  # this tests that a deep copy was done
        self.assertAlmostEqual(bm.toarray()[0, 0], 1)

        bm.copyfrom(bm0, deep=False)
        bm.get_block(0, 0).data.fill(1.0)
        self.assertAlmostEqual(
            bm0.toarray()[0, 0], 1
        )  # this tests that a shallow copy was done
        self.assertAlmostEqual(bm.toarray()[0, 0], 1)

    def test_copyto(self):
        bm0 = self.basic_m.copy()
        coo = bm0.tocoo()
        coo.data.fill(1.0)
        csr = coo.tocsr()
        csc = coo.tocsc()
        self.assertFalse(np.allclose(coo.toarray(), self.dense))
        self.assertFalse(np.allclose(csr.toarray(), self.dense))
        self.assertFalse(np.allclose(csc.toarray(), self.dense))
        bm0.copyto(coo)
        bm0.copyto(csr)
        bm0.copyto(csc)
        self.assertTrue(np.allclose(coo.toarray(), self.dense))
        self.assertTrue(np.allclose(csr.toarray(), self.dense))
        self.assertTrue(np.allclose(csc.toarray(), self.dense))

        flat = np.ones((8, 8))
        bm0.copyto(flat)
        self.assertTrue(np.allclose(flat, self.dense))

        bm = bm0.copy_structure()
        bm0.copyto(bm)
        self.assertTrue(np.allclose(bm.toarray(), self.dense))

        bm.get_block(0, 0).data.fill(1.0)
        self.assertAlmostEqual(
            bm0.toarray()[0, 0], 2
        )  # this tests that a deep copy was done
        self.assertAlmostEqual(bm.toarray()[0, 0], 1)

        bm0.copyto(bm, deep=False)
        bm.get_block(0, 0).data.fill(1.0)
        self.assertAlmostEqual(
            bm0.toarray()[0, 0], 1
        )  # this tests that a shallow copy was done
        self.assertAlmostEqual(bm.toarray()[0, 0], 1)

    def test_copy(self):
        clone = self.basic_m.copy()
        self.assertTrue(np.allclose(clone.toarray(), self.dense))
        clone.get_block(0, 0).data.fill(1)
        self.assertAlmostEqual(clone.toarray()[0, 0], 1)
        self.assertAlmostEqual(self.basic_m.toarray()[0, 0], 2)

        bm = self.basic_m.copy()
        clone = bm.copy(deep=False)
        self.assertTrue(np.allclose(clone.toarray(), self.dense))
        clone.get_block(0, 0).data.fill(1)
        self.assertAlmostEqual(clone.toarray()[0, 0], 1)
        self.assertAlmostEqual(bm.toarray()[0, 0], 1)

    def test_iadd(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m.copy()
        A_dense += A_dense
        A_block += A_block

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        A_dense = self.basic_m.toarray()
        A_block = self.basic_m.copy()
        A_dense += A_dense
        A_block += A_block.tocoo()

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        A_dense = self.basic_m.toarray()
        A_block = self.basic_m.copy()
        A_block += 2 * A_block.tocoo()

        self.assertTrue(np.allclose(A_block.toarray(), 3 * A_dense))

        with self.assertRaises(Exception) as context:
            A_block += 1.0

    def test_isub(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        A_dense -= A_dense
        A_block -= A_block

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        A_dense -= A_dense
        A_block -= A_block.tocoo()

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        A_dense = self.basic_m.toarray()
        A_block = self.basic_m.copy()
        A_block -= 2 * A_block.tocoo()

        self.assertTrue(np.allclose(A_block.toarray(), -A_dense))

        with self.assertRaises(Exception) as context:
            A_block -= 1.0

    def test_imul(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        print(A_dense)
        print(A_block.toarray())
        A_dense *= 3
        print(A_dense)
        print(A_block.toarray())
        A_block *= 3.0
        print(A_dense)
        print(A_block.toarray())

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        with self.assertRaises(Exception) as context:
            A_block *= A_block

        with self.assertRaises(Exception) as context:
            A_block *= A_block.tocoo()

        with self.assertRaises(Exception) as context:
            A_block *= A_block.toarray()

    def test_itruediv(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m.copy()
        A_dense /= 3
        A_block /= 3.0

        self.assertTrue(np.allclose(A_block.toarray(), A_dense))

        with self.assertRaises(Exception) as context:
            A_block /= A_block

        with self.assertRaises(Exception) as context:
            A_block /= A_block.tocoo()

        with self.assertRaises(Exception) as context:
            A_block /= A_block.toarray()

    def test_truediv(self):
        A_dense = self.basic_m.toarray()
        A_block = self.basic_m
        B_block = A_block / 3.0
        self.assertTrue(np.allclose(B_block.toarray(), A_dense / 3.0))

        with self.assertRaises(Exception) as context:
            b = A_block / A_block

        with self.assertRaises(Exception) as context:
            b = A_block / A_block.tocoo()

        with self.assertRaises(Exception) as context:
            b = A_block / A_block.toarray()

        with self.assertRaises(Exception) as context:
            B_block = 3.0 / A_block

    def test_eq(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A_flat = self.basic_m.tocoo()
            A_block = self.basic_m

            A_bool_flat = A_flat == 2.0
            A_bool_block = A_block == 2.0

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat == A_flat
            A_bool_block = A_block == A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = 2.0 != A_flat
            A_bool_block = 2.0 != A_block
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

    def test_ne(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A_flat = self.basic_m.tocoo()
            A_block = self.basic_m

            A_bool_flat = A_flat != 2.0
            A_bool_block = A_block != 2.0
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = 2.0 != A_flat
            A_bool_block = 2.0 != A_block
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat != A_flat
            A_bool_block = A_block != A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

    def test_le(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A_flat = self.basic_m.tocoo()
            A_block = self.basic_m

            A_bool_flat = A_flat <= 2.0
            A_bool_block = A_block <= 2.0
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            # A_bool_flat = 2.0 <= A_flat
            # A_bool_block = 2.0 <= A_block
            # self.assertTrue(np.allclose(A_bool_flat.toarray(),
            #                             A_bool_block.toarray()))

            A_bool_flat = A_flat <= A_flat
            A_bool_block = A_block <= A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat <= 2 * A_flat
            A_bool_block = A_block <= 2 * A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = 2.0 >= A_flat
            A_bool_block = 2.0 >= A_block
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

    def test_lt(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A_flat = self.basic_m.tocoo()
            A_block = self.basic_m

            A_bool_flat = A_flat < 2.0
            A_bool_block = A_block < 2.0

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            # A_bool_flat = 2.0 <= A_flat
            # A_bool_block = 2.0 <= A_block
            # self.assertTrue(np.allclose(A_bool_flat.toarray(),
            #                             A_bool_block.toarray()))

            A_bool_flat = A_flat < A_flat
            A_bool_block = A_block < A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat < 2 * A_flat
            A_bool_block = A_block < 2 * A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = 2.0 > A_flat
            A_bool_block = 2.0 > A_block
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

    def test_ge(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A_flat = self.basic_m.tocoo()
            A_block = self.basic_m

            A_bool_flat = A_flat >= 2.0
            A_bool_block = A_block >= 2.0
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = 2.0 <= A_flat
            A_bool_block = 2.0 <= A_block
            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat >= A_flat
            A_bool_block = A_block >= A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

            A_bool_flat = A_flat >= 0.5 * A_flat
            A_bool_block = A_block >= 0.5 * A_block

            self.assertTrue(np.allclose(A_bool_flat.toarray(), A_bool_block.toarray()))

    def test_gt(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            A = self.basic_m.copy()
            B = 2 * A.transpose()

            res = A > B
            expected = A.toarray() > B.toarray()
            self.assertTrue(np.allclose(res.toarray(), expected))

    def test_abs(self):
        row = np.array([0, 3, 1, 2, 3, 0])
        col = np.array([0, 0, 1, 2, 3, 3])
        data = -1.0 * np.array([2.0, 1, 3, 4, 5, 1])
        m = coo_matrix((data, (row, col)), shape=(4, 4))

        self.block_m = m

        bm = BlockMatrix(2, 2)
        bm.set_block(0, 0, m)
        bm.set_block(1, 1, m)
        bm.set_block(0, 1, m)

        abs_flat = abs(bm.tocoo())
        abs_mat = abs(bm)

        self.assertIsInstance(abs_mat, BlockMatrix)
        self.assertTrue(np.allclose(abs_flat.toarray(), abs_mat.toarray()))

    def test_getcol(self):
        m = self.basic_m

        flat_mat = m.tocoo()
        flat_col = flat_mat.getcol(2)
        block_col = m.getcol(2)
        self.assertTrue(np.allclose(flat_col.toarray().flatten(), block_col.flatten()))

        flat_col = flat_mat.getcol(4)
        block_col = m.getcol(4)
        self.assertTrue(np.allclose(flat_col.toarray().flatten(), block_col.flatten()))

        flat_col = flat_mat.getcol(6)
        block_col = m.getcol(6)
        self.assertTrue(np.allclose(flat_col.toarray().flatten(), block_col.flatten()))

    def test_getrow(self):
        m = self.basic_m

        flat_mat = m.tocoo()
        flat_row = flat_mat.getrow(2)
        block_row = m.getrow(2)
        self.assertTrue(np.allclose(flat_row.toarray().flatten(), block_row.flatten()))

        flat_row = flat_mat.getrow(7)
        block_row = m.getrow(7)
        self.assertTrue(np.allclose(flat_row.toarray().flatten(), block_row.flatten()))

    def test_nonzero(self):
        m = self.basic_m
        flat_mat = m.tocoo()
        flat_row, flat_col = flat_mat.nonzero()
        with self.assertRaises(Exception) as context:
            block_row, block_col = m.nonzero()

    def test_get_block_column_index(self):
        m = BlockMatrix(2, 4)
        m.set_block(0, 0, coo_matrix((3, 2)))
        m.set_block(0, 1, coo_matrix((3, 4)))
        m.set_block(0, 2, coo_matrix((3, 3)))
        m.set_block(0, 3, coo_matrix((3, 6)))
        m.set_block(1, 3, coo_matrix((5, 6)))

        bcol = m.get_block_column_index(8)
        self.assertEqual(bcol, 2)
        bcol = m.get_block_column_index(5)
        self.assertEqual(bcol, 1)
        bcol = m.get_block_column_index(14)
        self.assertEqual(bcol, 3)

    def test_get_block_row_index(self):
        m = BlockMatrix(2, 4)
        m.set_block(0, 0, coo_matrix((3, 2)))
        m.set_block(0, 1, coo_matrix((3, 4)))
        m.set_block(0, 2, coo_matrix((3, 3)))
        m.set_block(0, 3, coo_matrix((3, 6)))
        m.set_block(1, 3, coo_matrix((5, 6)))

        brow = m.get_block_row_index(0)
        self.assertEqual(brow, 0)
        brow = m.get_block_row_index(6)
        self.assertEqual(brow, 1)

    def test_matrix_multiply(self):
        """
        Test

        [A  B  C   *  [G  J   = [A*G + B*H + C*I    A*J + B*K + C*L
         D  E  F]      H  K      D*G + E*H + F*I    D*J + E*K + F*L]
                       I  L]
        """
        np.random.seed(0)
        A = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        B = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        C = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        D = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        E = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        F = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        G = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        H = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        I = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        J = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        K = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))
        L = sp.csr_matrix(np.random.normal(0, 10, (2, 2)))

        bm1 = BlockMatrix(2, 3)
        bm2 = BlockMatrix(3, 2)

        bm1.set_block(0, 0, A)
        bm1.set_block(0, 1, B)
        bm1.set_block(0, 2, C)
        bm1.set_block(1, 0, D)
        bm1.set_block(1, 1, E)
        bm1.set_block(1, 2, F)

        bm2.set_block(0, 0, G)
        bm2.set_block(1, 0, H)
        bm2.set_block(2, 0, I)
        bm2.set_block(0, 1, J)
        bm2.set_block(1, 1, K)
        bm2.set_block(2, 1, L)

        got = (bm1 * bm2).toarray()
        exp00 = (A * G + B * H + C * I).toarray()
        exp01 = (A * J + B * K + C * L).toarray()
        exp10 = (D * G + E * H + F * I).toarray()
        exp11 = (D * J + E * K + F * L).toarray()
        exp = np.zeros((4, 4))
        exp[0:2, 0:2] = exp00
        exp[0:2, 2:4] = exp01
        exp[2:4, 0:2] = exp10
        exp[2:4, 2:4] = exp11

        self.assertTrue(np.allclose(got, exp))

    def test_dimensions(self):
        bm = BlockMatrix(2, 2)
        self.assertTrue(bm.has_undefined_row_sizes())
        self.assertTrue(bm.has_undefined_col_sizes())
        with self.assertRaises(NotFullyDefinedBlockMatrixError):
            shape = bm.shape
        with self.assertRaises(NotFullyDefinedBlockMatrixError):
            bm.set_block(0, 0, BlockMatrix(2, 2))
        with self.assertRaises(NotFullyDefinedBlockMatrixError):
            row_sizes = bm.row_block_sizes()
        with self.assertRaises(NotFullyDefinedBlockMatrixError):
            col_sizes = bm.col_block_sizes()
        bm2 = BlockMatrix(2, 2)
        bm2.set_block(0, 0, coo_matrix((2, 2)))
        bm2.set_block(1, 1, coo_matrix((2, 2)))
        bm3 = bm2.copy()
        bm.set_block(0, 0, bm2)
        bm.set_block(1, 1, bm3)
        self.assertFalse(bm.has_undefined_row_sizes())
        self.assertFalse(bm.has_undefined_col_sizes())
        self.assertEqual(bm.shape, (8, 8))
        bm.set_block(0, 0, None)
        self.assertFalse(bm.has_undefined_row_sizes())
        self.assertFalse(bm.has_undefined_col_sizes())
        self.assertEqual(bm.shape, (8, 8))
        self.assertTrue(np.all(bm.row_block_sizes() == np.ones(2) * 4))
        self.assertTrue(np.all(bm.col_block_sizes() == np.ones(2) * 4))
        self.assertTrue(np.all(bm.row_block_sizes(copy=False) == np.ones(2) * 4))
        self.assertTrue(np.all(bm.col_block_sizes(copy=False) == np.ones(2) * 4))

    def test_transpose_with_empty_rows(self):
        m = BlockMatrix(2, 2)
        m.set_row_size(0, 2)
        m.set_row_size(1, 2)
        m.set_col_size(0, 2)
        m.set_col_size(1, 2)
        mt = m.transpose()
        self.assertEqual(mt.get_row_size(0), 2)
        self.assertEqual(mt.get_row_size(1), 2)
        self.assertEqual(mt.get_col_size(0), 2)
        self.assertEqual(mt.get_col_size(1), 2)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet
from pyomo.core.expr import identify_variables
from pyomo.environ import Constraint, value


def value_no_exception(c, div0=None):
    """
    Get value and ignore most exceptions (including division by 0).

    Args:
        c: a Pyomo component to get the value of
    Returns:
        A value, could be None
    """
    try:
        return value(c, exception=False)
    except ZeroDivisionError:
        return div0


def get_residual(ui_data, c):
    """
    Calculate the residual (constraint violation) of a constraint. This residual
    is always positive.

    Args:
        ui_data (UIData): user interface data, includes the cache for calculated
            values of the constraint body. This function uses the cached values
            and will not trigger recalculation. If variable values have changed,
            this may not yield accurate results.
        c(ConstraintData): a constraint or constraint data
    Returns:
        (float) residual
    """
    if c.upper is None:
        ub = None  # This is no upper bound
    else:
        ub = value_no_exception(c.upper, "Divide_by_0")
        if ub is None or isinstance(ub, str):
            # This is calc error
            return ub
    if c.lower is None:
        lb = None  # This is no lower bound
    else:
        lb = value_no_exception(c.lower, "Divide_by_0")
        if lb is None or isinstance(lb, str):
            # This is calc error
            return lb
    try:
        v = ui_data.value_cache[c]
    except KeyError:
        return None
    if v is None or isinstance(v, str):
        return v
    if lb is not None and v < lb:
        return lb - v
    if ub is not None and v > ub:
        return v - ub
    return 0.0


def active_equalities(blk):
    """
    Generator returning active equality constraints in a model.

    Args:
        blk: a Pyomo block in which to look for variables.
    """
    for o in blk.component_data_objects(Constraint, active=True):
        try:
            u = value(o.upper, exception=False)
            l = value(o.lower, exception=False)
            if u == l and l is not None:
                yield o
        except ZeroDivisionError:
            pass


def active_constraint_set(blk):
    """
    Return a set of active constraints in a model.

    Args:
        blk: a Pyomo block in which to look for constraints.
    Returns:
        (ComponentSet): Active equality constraints
    """
    return ComponentSet(blk.component_data_objects(Constraint, active=True))


def active_equality_set(blk):
    """
    Return a set of active equalities.

    Args:
        blk: a Pyomo block in which to look for variables.
    Returns:
        (ComponentSet): Active constraints
    """
    return ComponentSet(active_equalities(blk))


def count_free_variables(blk):
    """
    Count free variables that are in active equality constraints.  Ignore
    inequalities, because this is used in the degrees of freedom calculations
    """
    return len(free_variables_in_active_equalities_set(blk))


def count_equality_constraints(blk):
    """
    Count active equality constraints.
    """
    return len(active_equality_set(blk))


def count_constraints(blk):
    """
    Count active constraints.
    """
    return len(active_constraint_set(blk))


def degrees_of_freedom(blk):
    """
    Return the degrees of freedom.

    Args:
        blk (Block or BlockData): Block to count degrees of freedom in
    Returns:
        (int): Number of degrees of freedom
    """
    return count_free_variables(blk) - count_equality_constraints(blk)


def free_variables_in_active_equalities_set(blk):
    """
    Return a set of variables that are continued in active equalities.
    """
    vin = ComponentSet()
    for c in active_equalities(blk):
        for v in identify_variables(c.body):
            if not v.fixed:
                vin.add(v)
    return vin



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Param,
    Var,
    Objective,
    Constraint,
    Block,
    NonNegativeReals,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_block(_BaseTestModel):
    """
    A continuous linear model with nested blocks
    """

    description = "LP_block"
    capabilities = set(['linear'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.b = Block()
        model.B = Block([1, 2, 3])
        model.a = Param(initialize=1.0, mutable=True)
        model.b.x = Var(within=NonNegativeReals)
        model.B[1].x = Var(within=NonNegativeReals)

        model.obj = Objective(expr=model.b.x + 3.0 * model.B[1].x)
        model.obj.deactivate()
        model.B[2].c = Constraint(expr=-model.B[1].x <= -model.a)
        model.B[2].obj = Objective(expr=model.b.x + 3.0 * model.B[1].x + 2)
        model.B[3].c = Constraint(expr=(2.0, model.b.x / model.a - model.B[1].x, 10))

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.b.x.value = 1.0
        model.B[1].x.value = 1.0


@register_model
class LP_block_kernel(LP_block):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.b = pmo.block()
        model.B = pmo.block_dict((i, pmo.block()) for i in range(1, 4))
        model.a = pmo.parameter(value=1.0)
        model.b.x = pmo.variable(lb=0)
        model.B[1].x = pmo.variable(lb=0)

        model.obj = pmo.objective(expr=model.b.x + 3.0 * model.B[1].x)
        model.obj.deactivate()
        model.B[2].c = pmo.constraint(expr=-model.B[1].x <= -model.a)
        model.B[2].obj = pmo.objective(expr=model.b.x + 3.0 * model.B[1].x + 2)
        model.B[3].c = pmo.constraint(
            expr=(2.0, model.b.x / model.a - model.B[1].x, 10)
        )
